"""Functions to implement the randomized optimization and search algorithms."""

# Authors: Genevieve Hayes (modified by Andrew Rollings, Kyle Nakamura)
# License: BSD 3 clause

import numpy as np
from typing import Callable, Any
from mlrose_ky.decorators import short_name


@short_name("hc")
def hill_climb(
    problem: Any,
    max_iters: int = np.inf,
    restarts: int = 0,
    init_state: np.ndarray = None,
    curve: bool = False,
    random_state: int = None,
    state_fitness_callback: Callable = None,
    callback_user_info: Any = None,
) -> tuple[np.ndarray, float, np.ndarray | None]:
    """Use standard hill climbing to find the optimum for a given optimization problem.

    Parameters
    ----------
    problem: optimization object
        Object containing fitness function optimization problem to be solved.
        For example, `DiscreteOpt()`, `ContinuousOpt()` or `TSPOpt()`.
    max_iters: int, default: np.inf
        Maximum number of iterations of the algorithm for each restart.
    restarts: int, default: 0
        Number of random restarts.
    init_state: np.ndarray, default: None
        1-D Numpy array containing starting state for algorithm.
        If None, then a random state is used.
    curve: bool, default: False
        Boolean to keep fitness values for a curve.
        If False, then no curve is stored.
        If True, then a history of fitness values is provided as a
        third return value.
    random_state: int, default: None
        If random_state is a positive integer, random_state is the seed used
        by np.random.seed(); otherwise, the random seed is not set.
    state_fitness_callback: callable, default: None
        If specified, this callback will be invoked once per iteration.
        Parameters are (iteration, current best state, current best fit, user callback data).
        Return true to continue iterating, or false to stop.
    callback_user_info: any, default: None
        User data passed as last parameter of callback.

    Returns
    -------
    best_state: np.ndarray
        Numpy array containing state that optimizes the fitness function.
    best_fitness: float
        Value of fitness function at best state.
    fitness_curve: np.ndarray
        Numpy array containing the fitness at every iteration.
        Only returned if input argument curve is True.

    References
    ----------
    Russell, S. and P. Norvig (2010). *Artificial Intelligence: A Modern Approach*, 3rd edition.
    Prentice Hall, New Jersey, USA.
    """
    if not (isinstance(max_iters, int) or max_iters == np.inf or max_iters.is_integer()) or max_iters < 0:
        raise ValueError(f"max_iters must be a positive integer. Got {max_iters}")
    if not (isinstance(restarts, int) or restarts.is_integer()) or restarts < 0:
        raise ValueError(f"restarts must be a positive integer. Got {restarts}")
    if init_state is not None and len(init_state) != problem.get_length():
        raise ValueError(
            f"init_state must have the same length as the problem. " f"Expected length {problem.get_length()}, got {len(init_state)}"
        )

    # Set random seed
    if isinstance(random_state, int) and random_state > 0:
        np.random.seed(random_state)

    best_fitness = -np.inf
    best_state = None

    fitness_curve = []
    best_fitness_curve = []

    continue_iterating = True
    for current_restart in range(restarts + 1):
        # Initialize optimization problem
        if init_state is None:
            problem.reset()
        else:
            problem.set_state(init_state)

        callback_extra_data = None
        if state_fitness_callback is not None:
            callback_extra_data = callback_user_info + [("current_restart", current_restart)]
            # initial call with base data
            state_fitness_callback(
                iteration=0, state=problem.get_state(), fitness=problem.get_adjusted_fitness(), user_data=callback_extra_data
            )
        iters = 0
        while iters < max_iters:
            iters += 1

            # Find neighbors and determine best neighbor
            problem.find_neighbors()
            next_state = problem.best_neighbor()
            next_fitness = problem.eval_fitness(next_state)

            if curve:
                fitness_curve.append((problem.get_adjusted_fitness(), problem.fitness_evaluations))

            # invoke callback
            if state_fitness_callback is not None:
                max_attempts_reached = (iters == max_iters) or problem.can_stop()
                continue_iterating = state_fitness_callback(
                    iteration=iters,
                    attempt=None,
                    done=max_attempts_reached,
                    state=problem.get_state(),
                    fitness=problem.get_adjusted_fitness(),
                    curve=np.asarray(fitness_curve) if curve else None,
                    user_data=callback_extra_data,
                )
                # break out if requested
                if not continue_iterating:
                    break

            # If best neighbor is an improvement, move to that state
            current_fitness = problem.get_fitness()
            if next_fitness > current_fitness:
                problem.set_state(next_state)
            else:
                break

        # Update best state and best fitness
        if problem.get_fitness() > best_fitness:
            best_fitness = problem.get_fitness()
            best_state = problem.get_state()
            if curve:
                best_fitness_curve = [*fitness_curve]
                fitness_curve = []

        # break out if requested
        if not continue_iterating:
            break

    best_fitness = problem.get_maximize() * best_fitness

    if curve:
        return best_state, best_fitness, np.asarray(best_fitness_curve)

    return best_state, best_fitness, None



"""Unit tests for generators/"""

import pytest
import numpy as np

try:
    import mlrose_ky
except ImportError:
    import sys

    sys.path.append("..")
    import mlrose_ky

from mlrose_ky import ContinuousPeaks, DiscreteOpt, FourPeaks, SixPeaks, FlipFlopOpt, QueensOpt, OneMax
from mlrose_ky.generators import (
    ContinuousPeaksGenerator,
    FlipFlopGenerator,
    FourPeaksGenerator,
    SixPeaksGenerator,
    KnapsackGenerator,
    MaxKColorGenerator,
    QueensGenerator,
    TSPGenerator,
    OneMaxGenerator,
)

SEED = 12


# noinspection PyTypeChecker
class TestContinuousPeaksGenerator:

    def test_generate_size_zero(self):
        """Test generate method raises ValueError when size is zero"""
        with pytest.raises(ValueError) as excinfo:
            ContinuousPeaksGenerator.generate(SEED, size=0)
        assert str(excinfo.value) == "Size must be a positive integer. Got 0"

    def test_generate_negative_size(self):
        """Test generate method raises ValueError when size is a negative integer"""
        with pytest.raises(ValueError) as excinfo:
            ContinuousPeaksGenerator.generate(SEED, size=-5)
        assert str(excinfo.value) == "Size must be a positive integer. Got -5"

    def test_generate_non_integer_size(self):
        """Test generate method raises ValueError when size is a non-integer value"""
        with pytest.raises(ValueError) as excinfo:
            ContinuousPeaksGenerator.generate(SEED, size="ten")
        assert str(excinfo.value) == "Size must be a positive integer. Got ten"

    def test_generate_negative_threshold_percentage(self):
        """Test generate method raises ValueError when threshold_percentage is less than 0"""
        with pytest.raises(ValueError) as excinfo:
            ContinuousPeaksGenerator.generate(SEED, size=10, threshold_percentage=-0.1)
        assert str(excinfo.value) == "Threshold percentage must be between 0 and 1. Got -0.1"

    def test_generate_threshold_percentage_greater_than_one(self):
        """Test generate method raises ValueError when threshold_percentage is greater than 1"""
        with pytest.raises(ValueError) as excinfo:
            ContinuousPeaksGenerator.generate(SEED, size=10, threshold_percentage=1.1)
        assert str(excinfo.value) == "Threshold percentage must be between 0 and 1. Got 1.1"

    def test_generate_non_float_threshold_percentage(self):
        """Test generate method raises ValueError when threshold_percentage is a non-float value"""
        with pytest.raises(ValueError) as excinfo:
            ContinuousPeaksGenerator.generate(SEED, size=10, threshold_percentage="high")
        assert str(excinfo.value) == "Threshold percentage must be a float. Got str"

    def test_generate_default_size(self):
        """Test generate method with default size"""
        problem = ContinuousPeaksGenerator.generate(SEED)

        assert problem.length == 20

    def test_generate_default_threshold_percentage(self):
        """Test generate method with default threshold_percentage"""
        size = 20
        problem = ContinuousPeaksGenerator.generate(SEED, size)

        assert problem.length == size
        assert problem.fitness_fn.threshold_percentage == 0.1

    def test_generate_with_seed(self):
        """Test generate method with a specified SEED"""
        problem = ContinuousPeaksGenerator.generate(SEED)
        np.random.seed(SEED)
        expected_fitness = ContinuousPeaks()
        expected_problem = DiscreteOpt(length=20, fitness_fn=expected_fitness)

        assert problem.length == expected_problem.length
        assert problem.fitness_fn.__class__ == expected_problem.fitness_fn.__class__
        assert problem.fitness_fn.threshold_percentage == expected_problem.fitness_fn.threshold_percentage

    def test_generate_custom_size_and_threshold(self):
        """Test generate method with custom size and threshold_percentage"""
        size = 30
        threshold_percentage = 0.2
        problem = ContinuousPeaksGenerator.generate(SEED, size=size, threshold_percentage=threshold_percentage)

        assert problem.length == size
        assert problem.fitness_fn.threshold_percentage == threshold_percentage


# noinspection PyTypeChecker
class TestFlipFlopGenerator:

    def test_generate_size_zero(self):
        """Test generate method raises ValueError when size is zero."""
        with pytest.raises(ValueError) as excinfo:
            FlipFlopGenerator.generate(SEED, size=0)
        assert str(excinfo.value) == "Size must be a positive integer. Got 0."

    def test_generate_negative_size(self):
        """Test generate method raises ValueError when size is negative."""
        with pytest.raises(ValueError) as excinfo:
            FlipFlopGenerator.generate(SEED, size=-5)
        assert str(excinfo.value) == "Size must be a positive integer. Got -5."

    def test_generate_non_integer_size(self):
        """Test generate method raises ValueError when size is a non-integer value."""
        with pytest.raises(TypeError):
            FlipFlopGenerator.generate(SEED, size="ten")

    def test_generate_default_size(self):
        """Test generate method with default size."""
        problem = FlipFlopGenerator.generate(SEED)

        assert problem.length == 20

    def test_generate_with_seed(self):
        """Test generate method with a specified SEED."""
        problem = FlipFlopGenerator.generate(SEED)
        np.random.seed(SEED)
        expected_problem = FlipFlopOpt(length=20)

        assert problem.length == expected_problem.length
        assert problem.__class__ == expected_problem.__class__

    def test_generate_custom_size(self):
        """Test generate method with custom size."""
        size = 30
        problem = FlipFlopGenerator.generate(SEED, size=size)

        assert problem.length == size


# noinspection PyTypeChecker
class TestFourPeaksGenerator:

    def test_generate_size_zero(self):
        """Test generate method raises ValueError when size is zero."""
        with pytest.raises(ValueError) as excinfo:
            FourPeaksGenerator.generate(SEED, size=0)
        assert str(excinfo.value) == "Size must be a positive integer. Got 0."

    def test_generate_negative_size(self):
        """Test generate method raises ValueError when size is negative."""
        with pytest.raises(ValueError) as excinfo:
            FourPeaksGenerator.generate(SEED, size=-5)
        assert str(excinfo.value) == "Size must be a positive integer. Got -5."

    def test_generate_non_integer_size(self):
        """Test generate method raises ValueError when size is a non-integer value."""
        with pytest.raises(TypeError):
            FourPeaksGenerator.generate(SEED, size="ten")

    def test_generate_negative_threshold_percentage(self):
        """Test generate method raises ValueError when threshold_percentage is less than 0."""
        with pytest.raises(ValueError) as excinfo:
            FourPeaksGenerator.generate(SEED, size=10, threshold_percentage=-0.1)
        assert str(excinfo.value) == "Threshold percentage must be between 0 and 1. Got -0.1."

    def test_generate_threshold_percentage_greater_than_one(self):
        """Test generate method raises ValueError when threshold_percentage is greater than 1."""
        with pytest.raises(ValueError) as excinfo:
            FourPeaksGenerator.generate(SEED, size=10, threshold_percentage=1.1)
        assert str(excinfo.value) == "Threshold percentage must be between 0 and 1. Got 1.1."

    def test_generate_non_float_threshold_percentage(self):
        """Test generate method raises ValueError when threshold_percentage is a non-float value."""
        with pytest.raises(TypeError):
            FourPeaksGenerator.generate(SEED, size=10, threshold_percentage="high")

    def test_generate_default_size(self):
        """Test generate method with default size."""
        problem = FourPeaksGenerator.generate(SEED)

        assert problem.length == 20

    def test_generate_default_threshold_percentage(self):
        """Test generate method with default threshold_percentage."""
        size = 20
        problem = FourPeaksGenerator.generate(SEED, size)

        assert problem.length == size
        assert problem.fitness_fn.threshold_percentage == 0.1

    def test_generate_with_seed(self):
        """Test generate method with a specified SEED."""
        problem = FourPeaksGenerator.generate(SEED)
        np.random.seed(SEED)
        expected_fitness = FourPeaks()
        expected_problem = DiscreteOpt(length=20, fitness_fn=expected_fitness)

        assert problem.length == expected_problem.length
        assert problem.fitness_fn.__class__ == expected_problem.fitness_fn.__class__
        assert problem.fitness_fn.threshold_percentage == expected_problem.fitness_fn.threshold_percentage

    def test_generate_custom_size_and_threshold(self):
        """Test generate method with custom size and threshold_percentage."""
        size = 30
        threshold_percentage = 0.2
        problem = FourPeaksGenerator.generate(SEED, size=size, threshold_percentage=threshold_percentage)

        assert problem.length == size
        assert problem.fitness_fn.threshold_percentage == threshold_percentage


# noinspection PyTypeChecker
class TestSixPeaksGenerator:

    def test_generate_size_zero(self):
        """Test generate method raises ValueError when size is zero."""
        with pytest.raises(ValueError) as excinfo:
            SixPeaksGenerator.generate(SEED, size=0)
        assert str(excinfo.value) == "Size must be a positive integer. Got 0."

    def test_generate_negative_size(self):
        """Test generate method raises ValueError when size is negative."""
        with pytest.raises(ValueError) as excinfo:
            SixPeaksGenerator.generate(SEED, size=-5)
        assert str(excinfo.value) == "Size must be a positive integer. Got -5."

    def test_generate_non_integer_size(self):
        """Test generate method raises ValueError when size is a non-integer value."""
        with pytest.raises(TypeError):
            SixPeaksGenerator.generate(SEED, size="ten")

    def test_generate_negative_threshold_percentage(self):
        """Test generate method raises ValueError when threshold_percentage is less than 0."""
        with pytest.raises(ValueError) as excinfo:
            SixPeaksGenerator.generate(SEED, size=10, threshold_percentage=-0.1)
        assert str(excinfo.value) == "Threshold percentage must be between 0 and 1. Got -0.1."

    def test_generate_threshold_percentage_greater_than_one(self):
        """Test generate method raises ValueError when threshold_percentage is greater than 1."""
        with pytest.raises(ValueError) as excinfo:
            SixPeaksGenerator.generate(SEED, size=10, threshold_percentage=1.1)
        assert str(excinfo.value) == "Threshold percentage must be between 0 and 1. Got 1.1."

    def test_generate_non_float_threshold_percentage(self):
        """Test generate method raises ValueError when threshold_percentage is a non-float value."""
        with pytest.raises(TypeError):
            SixPeaksGenerator.generate(SEED, size=10, threshold_percentage="high")

    def test_generate_default_size(self):
        """Test generate method with default size."""
        problem = SixPeaksGenerator.generate(SEED)

        assert problem.length == 20

    def test_generate_default_threshold_percentage(self):
        """Test generate method with default threshold_percentage."""
        size = 20
        problem = SixPeaksGenerator.generate(SEED, size)

        assert problem.length == size
        assert problem.fitness_fn.threshold_percentage == 0.1

    def test_generate_with_seed(self):
        """Test generate method with a specified SEED."""
        problem = SixPeaksGenerator.generate(SEED)
        np.random.seed(SEED)
        expected_fitness = SixPeaks()
        expected_problem = DiscreteOpt(length=20, fitness_fn=expected_fitness)

        assert problem.length == expected_problem.length
        assert problem.fitness_fn.__class__ == expected_problem.fitness_fn.__class__
        assert problem.fitness_fn.threshold_percentage == expected_problem.fitness_fn.threshold_percentage

    def test_generate_custom_size_and_threshold(self):
        """Test generate method with custom size and threshold_percentage."""
        size = 30
        threshold_percentage = 0.2
        problem = SixPeaksGenerator.generate(SEED, size=size, threshold_percentage=threshold_percentage)

        assert problem.length == size
        assert problem.fitness_fn.threshold_percentage == threshold_percentage


# noinspection PyTypeChecker
class TestKnapsackGenerator:

    def test_generate_valid_case(self):
        """Test generate method with valid parameters."""
        number_of_item_types = 10
        max_item_count = 5
        max_weight_per_item = 25
        max_value_per_item = 10
        max_weight_percentage = 0.6
        multiply_by_max_item_count = True

        problem = KnapsackGenerator.generate(
            seed=SEED,
            number_of_item_types=number_of_item_types,
            max_item_count=max_item_count,
            max_weight_per_item=max_weight_per_item,
            max_value_per_item=max_value_per_item,
            max_weight_percentage=max_weight_percentage,
            multiply_by_max_item_count=multiply_by_max_item_count,
        )

        assert problem.length == number_of_item_types
        assert problem.max_val == max_item_count

    def test_generate_invalid_seed(self):
        """Test generate method raises ValueError when SEED is not an integer."""
        with pytest.raises(ValueError) as excinfo:
            KnapsackGenerator.generate(seed="not_an_int")
        assert str(excinfo.value) == "Seed must be an integer. Got not_an_int"

    def test_generate_invalid_number_of_item_types(self):
        """Test generate method raises ValueError when number_of_item_types is invalid."""
        with pytest.raises(ValueError) as excinfo:
            KnapsackGenerator.generate(SEED, number_of_item_types=0)
        assert str(excinfo.value) == "Number of item types must be a positive integer. Got 0"

    def test_generate_invalid_max_item_count(self):
        """Test generate method raises ValueError when max_item_count is invalid."""
        with pytest.raises(ValueError) as excinfo:
            KnapsackGenerator.generate(SEED, max_item_count=-1)
        assert str(excinfo.value) == "Max item count must be a positive integer. Got -1"

    def test_generate_invalid_max_weight_per_item(self):
        """Test generate method raises ValueError when max_weight_per_item is invalid."""
        with pytest.raises(ValueError) as excinfo:
            KnapsackGenerator.generate(SEED, max_weight_per_item=0)
        assert str(excinfo.value) == "Max weight per item must be a positive integer. Got 0"

    def test_generate_invalid_max_value_per_item(self):
        """Test generate method raises ValueError when max_value_per_item is invalid."""
        with pytest.raises(ValueError) as excinfo:
            KnapsackGenerator.generate(SEED, max_value_per_item=-10)
        assert str(excinfo.value) == "Max value per item must be a positive integer. Got -10"

    def test_generate_default_parameters(self):
        """Test generate method with default parameters."""
        problem = KnapsackGenerator.generate(seed=SEED)

        assert problem.length == 10
        assert problem.max_val == 5

    def test_generate_invalid_max_weight_percentage(self):
        """Test generate method raises ValueError when max_weight_percentage is invalid."""
        with pytest.raises(ValueError) as excinfo:
            KnapsackGenerator.generate(SEED, max_weight_percentage=1.5)
        assert str(excinfo.value) == "Max weight percentage must be a float between 0 and 1. Got 1.5"

    def test_generate_invalid_multiply_by_max_item_count(self):
        """Test generate method raises ValueError when multiply_by_max_item_count is not a boolean."""
        with pytest.raises(ValueError) as excinfo:
            KnapsackGenerator.generate(SEED, multiply_by_max_item_count="yes")
        assert str(excinfo.value) == "multiply_by_max_item_count must be a boolean. Got yes"

    def test_generate_max_weight_percentage_zero(self):
        """Test generate method with max_weight_percentage set to 0"""
        max_weight_percentage = 0.0

        with pytest.raises(ValueError) as excinfo:
            KnapsackGenerator.generate(seed=SEED, max_weight_percentage=max_weight_percentage)
        assert str(excinfo.value) == "max_weight_pct must be greater than 0."


# noinspection PyTypeChecker
class TestMaxKColorGenerator:

    def test_generate_negative_max_colors(self):
        """Test generate method raises ValueError when max_colors is a negative integer."""
        with pytest.raises(ValueError) as excinfo:
            MaxKColorGenerator.generate(SEED, number_of_nodes=10, max_colors=-3)
        assert str(excinfo.value) == "Max colors must be a positive integer or None. Got -3"

    def test_generate_non_integer_max_colors(self):
        """Test generate method raises ValueError when max_colors is a non-integer value."""
        with pytest.raises(ValueError) as excinfo:
            MaxKColorGenerator.generate(SEED, number_of_nodes=10, max_connections_per_node=3, max_colors="five")
        assert str(excinfo.value) == "Max colors must be a positive integer or None. Got five"

    def test_generate_seed_float(self):
        """Test generate method raises ValueError when SEED is a float."""
        with pytest.raises(ValueError) as excinfo:
            MaxKColorGenerator.generate(seed=1.5)
        assert str(excinfo.value) == "Seed must be an integer. Got 1.5"

    def test_generate_float_number_of_nodes(self):
        """Test generate method raises ValueError when number_of_nodes is a float."""
        with pytest.raises(ValueError) as excinfo:
            MaxKColorGenerator.generate(SEED, number_of_nodes=10.5)
        assert str(excinfo.value) == "Number of nodes must be a positive integer. Got 10.5"

    def test_generate_max_connections_per_node_float(self):
        """Test generate method raises ValueError when max_connections_per_node is a float."""
        with pytest.raises(ValueError) as excinfo:
            MaxKColorGenerator.generate(SEED, number_of_nodes=10, max_connections_per_node=4.5)
        assert str(excinfo.value) == "Max connections per node must be a positive integer. Got 4.5"

    def test_generate_maximize_string(self):
        """Test generate method raises ValueError when maximize is a string."""
        with pytest.raises(ValueError) as excinfo:
            MaxKColorGenerator.generate(SEED, maximize="true")
        assert str(excinfo.value) == "Maximize must be a boolean. Got true"

    def test_generate_single_node_one_connection(self):
        """Test generate method with one node and up to one connection."""
        number_of_nodes = 1
        max_connections_per_node = 1
        problem = MaxKColorGenerator.generate(SEED, number_of_nodes=number_of_nodes, max_connections_per_node=max_connections_per_node)

        assert problem.length == number_of_nodes
        assert problem.source_graph.number_of_edges() == 0

    def test_generate_single_node_two_connections(self):
        """Test generate method with one node and up to two connections."""
        number_of_nodes = 1
        max_connections_per_node = 2
        problem = MaxKColorGenerator.generate(SEED, number_of_nodes=number_of_nodes, max_connections_per_node=max_connections_per_node)

        assert problem.length == number_of_nodes
        assert problem.source_graph.number_of_edges() == 0

    def test_generate_two_nodes_one_connection(self):
        """Test generate method with two nodes and up to one connection."""
        number_of_nodes = 2
        max_connections_per_node = 1
        problem = MaxKColorGenerator.generate(SEED, number_of_nodes=number_of_nodes, max_connections_per_node=max_connections_per_node)

        assert problem.length == number_of_nodes
        assert problem.source_graph.number_of_edges() == 1

    def test_generate_two_nodes_two_connections(self):
        """Test generate method with two nodes and up to two connections."""
        number_of_nodes = 2
        max_connections_per_node = 2
        problem = MaxKColorGenerator.generate(SEED, number_of_nodes=number_of_nodes, max_connections_per_node=max_connections_per_node)

        assert problem.length == number_of_nodes
        assert problem.source_graph.number_of_edges() == 1

    def test_generate_large_graph(self):
        """Test generate method with a large graph."""
        number_of_nodes = 100
        max_connections_per_node = 10
        problem = MaxKColorGenerator.generate(SEED, number_of_nodes=number_of_nodes, max_connections_per_node=max_connections_per_node)

        assert problem.length == number_of_nodes
        assert problem.source_graph.number_of_edges() > 0

    def test_generate_no_edges(self):
        """Test generate method with no possible connections."""
        with pytest.raises(ValueError) as excinfo:
            MaxKColorGenerator.generate(SEED, number_of_nodes=10, max_connections_per_node=0)
        assert str(excinfo.value) == "Max connections per node must be a positive integer. Got 0"

    def test_generate_max_colors_none(self):
        """Test generate method with max_colors set to None."""
        number_of_nodes = 5
        max_connections_per_node = 3
        problem = MaxKColorGenerator.generate(SEED, number_of_nodes=number_of_nodes, max_connections_per_node=max_connections_per_node)

        assert problem.length == number_of_nodes
        assert problem.max_val > 1

    def test_generate_zero_nodes(self):
        """Test generate method raises ValueError when number_of_nodes is zero."""
        with pytest.raises(ValueError) as excinfo:
            MaxKColorGenerator.generate(SEED, number_of_nodes=0)
        assert str(excinfo.value) == "Number of nodes must be a positive integer. Got 0"

    def test_generate_large_max_colors(self):
        """Test generate method with a large max_colors value."""
        number_of_nodes = 10
        max_connections_per_node = 3
        max_colors = 100
        problem = MaxKColorGenerator.generate(
            SEED, number_of_nodes=number_of_nodes, max_connections_per_node=max_connections_per_node, max_colors=max_colors
        )

        assert problem.length == number_of_nodes
        assert problem.max_val == max_colors

    def test_generate_large_max_connections(self):
        """Test generate method with large max_connections_per_node value."""
        number_of_nodes = 10
        max_connections_per_node = 9
        problem = MaxKColorGenerator.generate(SEED, number_of_nodes=number_of_nodes, max_connections_per_node=max_connections_per_node)

        assert problem.length == number_of_nodes
        assert problem.source_graph.number_of_edges() > 0


# noinspection PyTypeChecker
class TestQueensGenerator:

    def test_generate_size_zero(self):
        """Test generate method raises ValueError when size is zero."""
        with pytest.raises(ValueError) as excinfo:
            QueensGenerator.generate(SEED, size=0)
        assert str(excinfo.value) == "Size must be a positive integer. Got 0"

    def test_generate_negative_size(self):
        """Test generate method raises ValueError when size is negative."""
        with pytest.raises(ValueError) as excinfo:
            QueensGenerator.generate(SEED, size=-5)
        assert str(excinfo.value) == "Size must be a positive integer. Got -5"

    def test_generate_non_integer_size(self):
        """Test generate method raises ValueError when size is a non-integer value."""
        with pytest.raises(ValueError) as excinfo:
            QueensGenerator.generate(SEED, size="ten")
        assert str(excinfo.value) == "Size must be a positive integer. Got ten"

    def test_generate_invalid_seed(self):
        """Test generate method raises ValueError when seed is not an integer."""
        with pytest.raises(ValueError) as excinfo:
            QueensGenerator.generate(seed="not_an_int")
        assert str(excinfo.value) == "Seed must be an integer. Got not_an_int"

    def test_generate_invalid_maximize(self):
        """Test generate method raises ValueError when maximize is not a boolean."""
        with pytest.raises(ValueError) as excinfo:
            QueensGenerator.generate(SEED, maximize="yes")
        assert str(excinfo.value) == "Maximize must be a boolean. Got yes"

    def test_generate_default_size(self):
        """Test generate method with default size."""
        problem = QueensGenerator.generate(SEED)

        assert problem.length == 20

    def test_generate_with_seed(self):
        """Test generate method with a specified seed."""
        problem = QueensGenerator.generate(SEED)
        np.random.seed(SEED)
        expected_problem = QueensOpt(length=20)

        assert problem.length == expected_problem.length
        assert problem.__class__ == expected_problem.__class__

    def test_generate_custom_size(self):
        """Test generate method with custom size."""
        size = 30
        problem = QueensGenerator.generate(SEED, size=size)

        assert problem.length == size


# noinspection PyTypeChecker
class TestTSPGenerator:

    def test_generate_invalid_seed(self):
        """Test generate method raises ValueError when seed is not an integer."""
        with pytest.raises(ValueError) as excinfo:
            TSPGenerator.generate(seed="not_an_int", number_of_cities=5)
        assert str(excinfo.value) == "Seed must be an integer. Got not_an_int"

    def test_generate_invalid_number_of_cities(self):
        """Test generate method raises ValueError when number_of_cities is not a positive integer."""
        with pytest.raises(ValueError) as excinfo:
            TSPGenerator.generate(SEED, number_of_cities=0)
        assert str(excinfo.value) == "Number of cities must be a positive integer. Got 0"

    def test_generate_invalid_area_width(self):
        """Test generate method raises ValueError when area_width is not a positive integer."""
        with pytest.raises(ValueError) as excinfo:
            TSPGenerator.generate(SEED, number_of_cities=5, area_width=0)
        assert str(excinfo.value) == "Area width must be a positive integer. Got 0"

    def test_generate_invalid_area_height(self):
        """Test generate method raises ValueError when area_height is not a positive integer."""
        with pytest.raises(ValueError) as excinfo:
            TSPGenerator.generate(SEED, number_of_cities=5, area_height=0)
        assert str(excinfo.value) == "Area height must be a positive integer. Got 0"

    def test_generate_default_parameters(self):
        """Test generate method with default parameters."""
        num_cities = 5
        problem = TSPGenerator.generate(seed=SEED, number_of_cities=num_cities)

        assert problem.length == num_cities
        assert problem.coords is not None
        assert problem.distances is not None
        assert problem.source_graph is not None

    def test_generate_custom_parameters(self):
        """Test generate method with custom parameters."""
        num_cities = 5
        problem = TSPGenerator.generate(seed=SEED, number_of_cities=num_cities, area_width=100, area_height=100)

        assert problem.length == num_cities
        assert problem.coords is not None
        assert problem.distances is not None
        assert problem.source_graph is not None

    def test_generate_no_duplicate_coordinates(self):
        """Test generate method ensures no duplicate coordinates."""
        problem = TSPGenerator.generate(seed=SEED, number_of_cities=5)
        coords = problem.coords

        assert len(coords) == len(set(coords))

    def test_generate_distances(self):
        """Test generate method calculates distances correctly."""
        problem = TSPGenerator.generate(seed=SEED, number_of_cities=5)
        distances = problem.distances
        for u, v, d in distances:
            assert d == np.linalg.norm(np.subtract(problem.coords[u], problem.coords[v]))

    def test_generate_graph(self):
        """Test generate method creates a valid graph."""
        problem = TSPGenerator.generate(seed=SEED, number_of_cities=5)
        graph = problem.source_graph

        assert graph.number_of_nodes() == 5
        assert graph.number_of_edges() == len(problem.distances)

    def test_generate_valid_case(self):
        """Test generate method with valid parameters."""
        number_of_item_types = 10
        max_item_count = 5
        max_weight_per_item = 25
        max_value_per_item = 10
        max_weight_percentage = 0.6
        multiply_by_max_item_count = True

        problem = KnapsackGenerator.generate(
            seed=SEED,
            number_of_item_types=number_of_item_types,
            max_item_count=max_item_count,
            max_weight_per_item=max_weight_per_item,
            max_value_per_item=max_value_per_item,
            max_weight_percentage=max_weight_percentage,
            multiply_by_max_item_count=multiply_by_max_item_count,
        )

        assert problem.length == number_of_item_types
        assert problem.max_val == max_item_count

    def test_generate_invalid_number_of_item_types(self):
        """Test generate method raises ValueError when number_of_item_types is invalid."""
        with pytest.raises(ValueError) as excinfo:
            KnapsackGenerator.generate(SEED, number_of_item_types=0)
        assert str(excinfo.value) == "Number of item types must be a positive integer. Got 0"

    def test_generate_invalid_max_item_count(self):
        """Test generate method raises ValueError when max_item_count is invalid."""
        with pytest.raises(ValueError) as excinfo:
            KnapsackGenerator.generate(SEED, max_item_count=-1)
        assert str(excinfo.value) == "Max item count must be a positive integer. Got -1"

    def test_generate_invalid_max_weight_per_item(self):
        """Test generate method raises ValueError when max_weight_per_item is invalid."""
        with pytest.raises(ValueError) as excinfo:
            KnapsackGenerator.generate(SEED, max_weight_per_item=0)
        assert str(excinfo.value) == "Max weight per item must be a positive integer. Got 0"

    def test_generate_invalid_max_value_per_item(self):
        """Test generate method raises ValueError when max_value_per_item is invalid."""
        with pytest.raises(ValueError) as excinfo:
            KnapsackGenerator.generate(SEED, max_value_per_item=-10)
        assert str(excinfo.value) == "Max value per item must be a positive integer. Got -10"

    def test_generate_invalid_max_weight_percentage(self):
        """Test generate method raises ValueError when max_weight_percentage is invalid."""
        with pytest.raises(ValueError) as excinfo:
            KnapsackGenerator.generate(SEED, max_weight_percentage=1.5)
        assert str(excinfo.value) == "Max weight percentage must be a float between 0 and 1. Got 1.5"

    def test_generate_invalid_multiply_by_max_item_count(self):
        """Test generate method raises ValueError when multiply_by_max_item_count is not a boolean."""
        with pytest.raises(ValueError) as excinfo:
            KnapsackGenerator.generate(SEED, multiply_by_max_item_count="yes")
        assert str(excinfo.value) == "multiply_by_max_item_count must be a boolean. Got yes"

    def test_generate_max_weight_percentage_zero(self):
        """Test generate method with max_weight_percentage set to 0"""
        with pytest.raises(ValueError) as excinfo:
            KnapsackGenerator.generate(seed=SEED, max_weight_percentage=0.0)
        assert str(excinfo.value) == "max_weight_pct must be greater than 0."


# noinspection PyTypeChecker
class TestOneMaxGenerator:

    def test_generate_size_zero(self):
        """Test generate method raises ValueError when size is zero."""
        with pytest.raises(ValueError) as excinfo:
            OneMaxGenerator.generate(SEED, size=0)
        assert str(excinfo.value) == "Size must be a positive integer. Got 0"

    def test_generate_negative_size(self):
        """Test generate method raises ValueError when size is negative."""
        with pytest.raises(ValueError) as excinfo:
            OneMaxGenerator.generate(SEED, size=-5)
        assert str(excinfo.value) == "Size must be a positive integer. Got -5"

    def test_generate_non_integer_size(self):
        """Test generate method raises ValueError when size is a non-integer value."""
        with pytest.raises(ValueError) as excinfo:
            OneMaxGenerator.generate(SEED, size="ten")
        assert str(excinfo.value) == "Size must be a positive integer. Got ten"

    def test_generate_invalid_seed(self):
        """Test generate method raises ValueError when seed is not an integer."""
        with pytest.raises(ValueError) as excinfo:
            OneMaxGenerator.generate(seed="not_an_int")
        assert str(excinfo.value) == "Seed must be an integer. Got not_an_int"

    def test_generate_default_size(self):
        """Test generate method with default size."""
        problem = OneMaxGenerator.generate(SEED)

        assert problem.length == 20

    def test_generate_with_seed(self):
        """Test generate method with a specified seed."""
        problem = OneMaxGenerator.generate(SEED)
        np.random.seed(SEED)
        expected_problem = DiscreteOpt(length=20, fitness_fn=OneMax())

        assert problem.length == expected_problem.length
        assert problem.__class__ == expected_problem.__class__

    def test_generate_custom_size(self):
        """Test generate method with custom size."""
        size = 30
        problem = OneMaxGenerator.generate(SEED, size=size)

        assert problem.length == size



"""Unit tests for neural/"""

# Author: Genevieve Hayes (modified by Kyle Nakamura)
# License: BSD 3 clause

import numpy as np
from sklearn.model_selection import StratifiedShuffleSplit, learning_curve

try:
    import mlrose_ky
except ImportError:
    import sys

    sys.path.append("..")

from mlrose_ky.neural.linear_regression import LinearRegression
from mlrose_ky.neural.logistic_regression import LogisticRegression
from mlrose_ky.neural.fitness.network_weights import NetworkWeights
from mlrose_ky.neural.neural_network import NeuralNetwork
from mlrose_ky.opt_probs import ContinuousOpt
from mlrose_ky import flatten_weights, unflatten_weights, identity, sigmoid, softmax
from mlrose_ky.algorithms.gd import gradient_descent


class TestNeural:
    """Tests for neural.py functions."""

    def test_flatten_weights(self):
        """Test flatten_weights function"""
        x = np.arange(12)
        y = np.arange(6)
        z = np.arange(16)

        a = np.reshape(x, (4, 3))
        b = np.reshape(y, (3, 2))
        c = np.reshape(z, (2, 8))

        weights = [a, b, c]
        flat = list(x) + list(y) + list(z)

        assert np.array_equal(np.array(flatten_weights(weights)), np.array(flat))

    def test_unflatten_weights(self):
        """Test unflatten_weights function"""
        x = np.arange(12)
        y = np.arange(6)
        z = np.arange(16)

        a = np.reshape(x, (4, 3))
        b = np.reshape(y, (3, 2))
        c = np.reshape(z, (2, 8))

        flat = list(x) + list(y) + list(z)
        nodes = [4, 3, 2, 8]
        weights = unflatten_weights(np.asarray(flat), nodes)

        assert np.array_equal(weights[0], a) and np.array_equal(weights[1], b) and np.array_equal(weights[2], c)

    def test_gradient_descent(self):
        """Test gradient_descent function"""
        X = np.array([[0, 1, 0, 1], [0, 0, 0, 0], [1, 1, 1, 1], [1, 1, 1, 1], [0, 0, 1, 1], [1, 0, 0, 0]])
        y = np.reshape(np.array([1, 1, 0, 0, 1, 1]), [6, 1])

        nodes = [4, 2, 1]
        fitness = NetworkWeights(X, y, nodes, activation=identity, bias=False, is_classifier=False)

        problem = ContinuousOpt(10, fitness, maximize=False, min_val=-1)

        test_weights = np.ones(10)
        test_fitness = -1 * problem.eval_fitness(test_weights)
        best_state, best_fitness, _ = gradient_descent(problem)

        assert len(best_state) == 10 and min(best_state) >= -1 and max(best_state) <= 1 and best_fitness < test_fitness

    def test_gradient_descent_iter1(self):
        """Test gradient_descent function gets the correct answer after a
        single iteration"""
        X = np.array([[0, 1, 0, 1], [0, 0, 0, 0], [1, 1, 1, 1], [1, 1, 1, 1], [0, 0, 1, 1], [1, 0, 0, 0]])
        y = np.reshape(np.array([1, 1, 0, 0, 1, 1]), [6, 1])

        nodes = [4, 2, 1]
        fitness = NetworkWeights(X, y, nodes, activation=identity, bias=False, is_classifier=False)

        problem = ContinuousOpt(10, fitness, maximize=False, min_val=-1)
        init_weights = np.ones(10)
        best_state, best_fitness, _ = gradient_descent(problem, max_iters=1, init_state=init_weights)
        x = np.array([-0.7, -0.7, -0.9, -0.9, -0.9, -0.9, -1, -1, -1, -1])

        assert np.allclose(best_state, x, atol=0.001) and round(best_fitness, 2) == 19.14


class TestNeuralWeights:
    """Tests for NeuralWeights class."""

    def test_evaluate_no_bias_classifier(self):
        """Test evaluate method for binary classifier with no bias term"""
        X = np.array([[0, 1, 0, 1], [0, 0, 0, 0], [1, 1, 1, 1], [1, 1, 1, 1], [0, 0, 1, 1], [1, 0, 0, 0]])
        y = np.reshape(np.array([1, 1, 0, 0, 1, 1]), [6, 1])

        nodes = [4, 2, 1]
        fitness = NetworkWeights(X, y, nodes, activation=identity, bias=False)

        a = list(np.arange(8) + 1)
        b = list(0.01 * (np.arange(2) + 1))
        weights = a + b

        assert round(fitness.evaluate(np.asarray(weights)), 4) == 0.7393

    def test_evaluate_no_bias_multi(self):
        """Test evaluate method for multivariate classifier with no bias term"""
        X = np.array([[0, 1, 0, 1], [0, 0, 0, 0], [1, 1, 1, 1], [1, 1, 1, 1], [0, 0, 1, 1], [1, 0, 0, 0]])
        y = np.array([[1, 1], [1, 0], [0, 0], [0, 0], [1, 0], [1, 1]])

        nodes = [4, 2, 2]
        fitness = NetworkWeights(X, y, nodes, activation=identity, bias=False)

        a = list(np.arange(8) + 1)
        b = list(0.01 * (np.arange(4) + 1))
        weights = a + b

        assert round(fitness.evaluate(np.asarray(weights)), 4) == 0.7183

    def test_evaluate_no_bias_regressor(self):
        """Test evaluate method for regressor with no bias term"""
        X = np.array([[0, 1, 0, 1], [0, 0, 0, 0], [1, 1, 1, 1], [1, 1, 1, 1], [0, 0, 1, 1], [1, 0, 0, 0]])
        y = np.reshape(np.array([1, 1, 0, 0, 1, 1]), [6, 1])

        nodes = [4, 2, 1]
        fitness = NetworkWeights(X, y, nodes, activation=identity, bias=False, is_classifier=False)

        a = list(np.arange(8) + 1)
        b = list(0.01 * (np.arange(2) + 1))
        weights = a + b

        assert round(fitness.evaluate(np.asarray(weights)), 4) == 0.5542

    def test_evaluate_bias_regressor(self):
        """Test evaluate method for regressor with bias term"""
        X = np.array([[0, 1, 0, 1], [0, 0, 0, 0], [1, 1, 1, 1], [1, 1, 1, 1], [0, 0, 1, 1], [1, 0, 0, 0]])
        y = np.reshape(np.array([1, 1, 0, 0, 1, 1]), [6, 1])

        nodes = [5, 2, 1]
        fitness = NetworkWeights(X, y, nodes, activation=identity, is_classifier=False)

        a = list(np.arange(10) + 1)
        b = list(0.01 * (np.arange(2) + 1))
        weights = a + b

        assert round(fitness.evaluate(np.asarray(weights)), 4) == 0.4363

    def test_calculate_updates(self):
        """Test calculate_updates method"""
        X = np.array([[0, 1, 0, 1], [0, 0, 0, 0], [1, 1, 1, 1], [1, 1, 1, 1], [0, 0, 1, 1], [1, 0, 0, 0]])
        y = np.reshape(np.array([1, 1, 0, 0, 1, 1]), [6, 1])

        nodes = [4, 2, 1]
        fitness = NetworkWeights(X, y, nodes, activation=identity, bias=False, is_classifier=False, learning_rate=1)

        a = list(np.arange(8) + 1)
        b = list(0.01 * (np.arange(2) + 1))
        weights = a + b
        fitness.evaluate(np.asarray(weights))

        updates = fitness.calculate_updates()
        update1 = np.array([[-0.0017, -0.0034], [-0.0046, -0.0092], [-0.0052, -0.0104], [0.0014, 0.0028]])
        update2 = np.array([[-3.17], [-4.18]])

        assert np.allclose(updates[0], update1, atol=0.001) and np.allclose(updates[1], update2, atol=0.001)


class TestNeuralNetwork:
    """Tests for NeuralNetwork class."""

    def test_fit_random_hill_climb(self):
        """Test fit method using the random hill climbing algorithm"""
        network = NeuralNetwork(hidden_nodes=[2], activation="identity", bias=False, learning_rate=1, clip_max=1, max_attempts=100)

        X = np.array([[0, 1, 0, 1], [0, 0, 0, 0], [1, 1, 1, 1], [1, 1, 1, 1], [0, 0, 1, 1], [1, 0, 0, 0]])
        y = np.reshape(np.array([1, 1, 0, 0, 1, 1]), [6, 1])

        weights = np.ones(10)
        network.fit(X, y, init_weights=weights)
        fitted = network.fitted_weights

        assert sum(fitted) < 10 and len(fitted) == 10 and min(fitted) >= -1 and max(fitted) <= 1

    def test_fit_simulated_annealing(self):
        """Test fit method using the simulated_annealing algorithm"""
        network = NeuralNetwork(
            hidden_nodes=[2],
            activation="identity",
            algorithm="simulated_annealing",
            bias=False,
            learning_rate=1,
            clip_max=1,
            max_attempts=100,
        )

        X = np.array([[0, 1, 0, 1], [0, 0, 0, 0], [1, 1, 1, 1], [1, 1, 1, 1], [0, 0, 1, 1], [1, 0, 0, 0]])
        y = np.reshape(np.array([1, 1, 0, 0, 1, 1]), [6, 1])

        weights = np.ones(10)
        network.fit(X, y, init_weights=weights)
        fitted = network.fitted_weights

        assert sum(fitted) < 10 and len(fitted) == 10 and min(fitted) >= -1 and max(fitted) <= 1

    def test_fit_genetic_alg(self):
        """Test fit method using the genetic_alg algorithm"""
        network = NeuralNetwork(
            hidden_nodes=[2], activation="identity", algorithm="genetic_alg", bias=False, learning_rate=1, clip_max=1, max_attempts=100
        )

        X = np.array([[0, 1, 0, 1], [0, 0, 0, 0], [1, 1, 1, 1], [1, 1, 1, 1], [0, 0, 1, 1], [1, 0, 0, 0]])
        y = np.reshape(np.array([1, 1, 0, 0, 1, 1]), [6, 1])

        network.fit(X, y)
        fitted = network.fitted_weights

        assert sum(fitted) < 10 and len(fitted) == 10 and min(fitted) >= -1 and max(fitted) <= 1

    def test_fit_gradient_descent(self):
        """Test fit method using the gradient_descent algorithm"""
        network = NeuralNetwork(
            hidden_nodes=[2], activation="identity", algorithm="gradient_descent", bias=False, learning_rate=1, clip_max=1, max_attempts=100
        )

        X = np.array([[0, 1, 0, 1], [0, 0, 0, 0], [1, 1, 1, 1], [1, 1, 1, 1], [0, 0, 1, 1], [1, 0, 0, 0]])
        y = np.reshape(np.array([1, 1, 0, 0, 1, 1]), [6, 1])

        weights = np.ones(10)
        network.fit(X, y, init_weights=weights)
        fitted = network.fitted_weights

        assert sum(fitted) < 10 and len(fitted) == 10 and min(fitted) >= -1 and max(fitted) <= 1

    def test_predict_no_bias(self):
        """Test predict method with no bias term"""
        network = NeuralNetwork(hidden_nodes=[2], activation="identity", bias=False, learning_rate=1, clip_max=1, max_attempts=100)

        X = np.array([[0, 1, 0, 1], [0, 0, 0, 0], [1, 1, 1, 1], [1, 1, 1, 1], [0, 0, 1, 1], [1, 0, 0, 0]])
        network.fitted_weights = np.array([0.2, 0.5, 0.3, 0.4, 0.4, 0.3, 0.5, 0.2, -1, 1, 1, -1])
        network.node_list = [4, 2, 2]
        network.output_activation = softmax

        probs = np.array([[0.40131, 0.59869], [0.5, 0.5], [0.5, 0.5], [0.5, 0.5], [0.31003, 0.68997], [0.64566, 0.35434]])
        labels = np.array([[0, 1], [1, 0], [1, 0], [1, 0], [0, 1], [1, 0]])

        assert np.array_equal(network.predict(X), labels) and np.allclose(network.predicted_probs, probs, atol=0.0001)

    def test_predict_bias(self):
        """Test predict method with bias term"""
        network = NeuralNetwork(hidden_nodes=[2], activation="identity", learning_rate=1, clip_max=1, max_attempts=100)

        X = np.array([[0, 1, 0, 1], [0, 0, 0, 0], [1, 1, 1, 1], [1, 1, 1, 1], [0, 0, 1, 1], [1, 0, 0, 0]])

        network.fitted_weights = np.array([0.2, 0.5, 0.3, 0.4, 0.4, 0.3, 0.5, 0.2, 1, -1, -0.1, 0.1, 0.1, -0.1])
        network.node_list = [5, 2, 2]
        network.output_activation = softmax

        probs = np.array(
            [[0.39174, 0.60826], [0.40131, 0.59869], [0.40131, 0.59869], [0.40131, 0.59869], [0.38225, 0.61775], [0.41571, 0.58419]]
        )
        labels = np.array([[0, 1], [0, 1], [0, 1], [0, 1], [0, 1], [0, 1]])

        assert np.array_equal(network.predict(X), labels) and np.allclose(network.predicted_probs, probs, atol=0.0001)

    def test_learning_curve(self):
        """Test scikit-learn learning curve method."""
        network = NeuralNetwork(
            hidden_nodes=[2],
            activation="identity",
            algorithm="simulated_annealing",
            curve=True,
            learning_rate=1,
            clip_max=1,
            max_attempts=100,
        )

        X = np.array(
            [
                [0, 1, 0, 1],
                [0, 0, 1, 0],
                [1, 1, 0, 1],
                [1, 0, 1, 1],
                [0, 0, 1, 1],
                [1, 0, 0, 0],
                [1, 1, 1, 0],
                [0, 1, 1, 0],
                [0, 0, 0, 1],
                [1, 1, 0, 0],
                [0, 1, 1, 1],
                [1, 0, 1, 0],
                [1, 1, 1, 1],
                [0, 0, 0, 0],
                [0, 1, 0, 0],
                [1, 0, 0, 1],
            ]
        )
        y = np.array([1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1, 0, 1, 0, 1, 0])

        train_sizes = [0.5, 1.0]
        cv = StratifiedShuffleSplit(n_splits=1, test_size=0.2, random_state=42)
        train_sizes, train_scores, test_scores = learning_curve(network, X, y, train_sizes=train_sizes, cv=cv, scoring="accuracy")

        assert not np.isnan(train_scores).any() and not np.isnan(test_scores).any()


class TestLinearRegression:
    """Tests for LinearRegression class."""

    def test_fit_random_hill_climb(self):
        """Test fit method using the random hill climbing algorithm"""
        network = LinearRegression(bias=False, learning_rate=1, clip_max=1, max_attempts=100)

        X = np.array([[0, 1, 0, 1], [0, 0, 0, 0], [1, 1, 1, 1], [1, 1, 1, 1], [0, 0, 1, 1], [1, 0, 0, 0]])
        y = np.reshape(np.array([1, 1, 0, 0, 1, 1]), [6, 1])

        weights = np.ones(4)
        network.fit(X, y, init_weights=weights)
        fitted = network.fitted_weights

        assert sum(fitted) < 4 and len(fitted) == 4 and min(fitted) >= -1 and max(fitted) <= 1

    def test_fit_simulated_annealing(self):
        """Test fit method using the simulated_annealing algorithm"""
        network = LinearRegression(algorithm="simulated_annealing", bias=False, learning_rate=1, clip_max=1, max_attempts=100)

        X = np.array([[0, 1, 0, 1], [0, 0, 0, 0], [1, 1, 1, 1], [1, 1, 1, 1], [0, 0, 1, 1], [1, 0, 0, 0]])
        y = np.reshape(np.array([1, 1, 0, 0, 1, 1]), [6, 1])

        weights = np.ones(4)
        network.fit(X, y, init_weights=weights)
        fitted = network.fitted_weights

        assert sum(fitted) < 4 and len(fitted) == 4 and min(fitted) >= -1 and max(fitted) <= 1

    def test_fit_genetic_alg(self):
        """Test fit method using the genetic_alg algorithm"""
        network = LinearRegression(algorithm="genetic_alg", bias=False, learning_rate=1, clip_max=1, max_attempts=100)

        X = np.array([[0, 1, 0, 1], [0, 0, 0, 0], [1, 1, 1, 1], [1, 1, 1, 1], [0, 0, 1, 1], [1, 0, 0, 0]])
        y = np.reshape(np.array([1, 1, 0, 0, 1, 1]), [6, 1])

        network.fit(X, y)
        fitted = network.fitted_weights

        assert sum(fitted) < 4 and len(fitted) == 4 and min(fitted) >= -1 and max(fitted) <= 1

    def test_fit_gradient_descent(self):
        """Test fit method using the gradient_descent algorithm"""
        network = LinearRegression(algorithm="gradient_descent", bias=False, clip_max=1, max_attempts=100)

        X = np.array([[0, 1, 0, 1], [0, 0, 0, 0], [1, 1, 1, 1], [1, 1, 1, 1], [0, 0, 1, 1], [1, 0, 0, 0]])
        y = np.reshape(np.array([1, 1, 0, 0, 1, 1]), [6, 1])

        weights = np.ones(4)
        network.fit(X, y, init_weights=weights)
        fitted = network.fitted_weights

        assert sum(fitted) <= 4 and len(fitted) == 4 and min(fitted) >= -1 and max(fitted) <= 1

    def test_predict_no_bias(self):
        """Test predict method with no bias term"""
        network = LinearRegression(bias=False, learning_rate=1, clip_max=1, max_attempts=100)

        X = np.array([[0, 1, 0, 1], [0, 0, 0, 0], [1, 1, 1, 1], [1, 1, 1, 1], [0, 0, 1, 1], [1, 0, 0, 0]])

        network.fitted_weights = np.ones(4)
        network.node_list = [4, 1]
        network.output_activation = identity

        x = np.reshape(np.array([2, 0, 4, 4, 2, 1]), [6, 1])

        assert np.array_equal(network.predict(X), x)

    def test_predict_bias(self):
        """Test predict method with bias term"""
        network = LinearRegression(learning_rate=1, clip_max=1, max_attempts=100)

        X = np.array([[0, 1, 0, 1], [0, 0, 0, 0], [1, 1, 1, 1], [1, 1, 1, 1], [0, 0, 1, 1], [1, 0, 0, 0]])

        network.fitted_weights = np.ones(5)
        network.node_list = [5, 1]
        network.output_activation = identity

        x = np.reshape(np.array([3, 1, 5, 5, 3, 2]), [6, 1])

        assert np.array_equal(network.predict(X), x)


class TestLogisticRegression:
    """Tests for LogisticRegression class."""

    def test_fit_random_hill_climb(self):
        """Test fit method using the random hill climbing algorithm"""
        network = LogisticRegression(bias=False, learning_rate=1, clip_max=1, max_attempts=100)

        X = np.array([[0, 1, 0, 1], [0, 0, 0, 0], [1, 1, 1, 1], [1, 1, 1, 1], [0, 0, 1, 1], [1, 0, 0, 0]])
        y = np.reshape(np.array([1, 1, 0, 0, 1, 1]), [6, 1])

        weights = np.ones(4)
        network.fit(X, y, init_weights=weights)
        fitted = network.fitted_weights

        assert sum(fitted) < 4 and len(fitted) == 4 and min(fitted) >= -1 and max(fitted) <= 1

    def test_fit_simulated_annealing(self):
        """Test fit method using the simulated_annealing algorithm"""
        network = LogisticRegression(algorithm="simulated_annealing", bias=False, learning_rate=1, clip_max=1, max_attempts=100)

        X = np.array([[0, 1, 0, 1], [0, 0, 0, 0], [1, 1, 1, 1], [1, 1, 1, 1], [0, 0, 1, 1], [1, 0, 0, 0]])
        y = np.reshape(np.array([1, 1, 0, 0, 1, 1]), [6, 1])

        weights = np.ones(4)
        network.fit(X, y, init_weights=weights)
        fitted = network.fitted_weights

        assert sum(fitted) < 4 and len(fitted) == 4 and min(fitted) >= -1 and max(fitted) <= 1

    def test_fit_genetic_alg(self):
        """Test fit method using the genetic_alg algorithm"""
        network = LogisticRegression(algorithm="genetic_alg", bias=False, learning_rate=1, clip_max=1, max_attempts=100)

        X = np.array([[0, 1, 0, 1], [0, 0, 0, 0], [1, 1, 1, 1], [1, 1, 1, 1], [0, 0, 1, 1], [1, 0, 0, 0]])
        y = np.reshape(np.array([1, 1, 0, 0, 1, 1]), [6, 1])

        network.fit(X, y)
        fitted = network.fitted_weights

        assert sum(fitted) < 4 and len(fitted) == 4 and min(fitted) >= -1 and max(fitted) <= 1

    def test_fit_gradient_descent(self):
        """Test fit method using the gradient_descent algorithm"""
        network = LogisticRegression(algorithm="gradient_descent", bias=False, clip_max=1, max_attempts=100)

        X = np.array([[0, 1, 0, 1], [0, 0, 0, 0], [1, 1, 1, 1], [1, 1, 1, 1], [0, 0, 1, 1], [1, 0, 0, 0]])
        y = np.reshape(np.array([1, 1, 0, 0, 1, 1]), [6, 1])

        weights = np.ones(4)
        network.fit(X, y, init_weights=weights)
        fitted = network.fitted_weights

        assert sum(fitted) <= 4 and len(fitted) == 4 and min(fitted) >= -1 and max(fitted) <= 1

    def test_predict_no_bias(self):
        """Test predict method with no bias term"""
        network = LogisticRegression(bias=False, learning_rate=1, clip_max=1, max_attempts=100)

        X = np.array([[0, 1, 0, 1], [0, 0, 0, 0], [1, 1, 1, 1], [1, 1, 1, 1], [0, 0, 1, 1], [1, 0, 0, 0]])

        network.fitted_weights = np.array([-1, 1, 1, 1])
        network.node_list = [4, 1]
        network.output_activation = sigmoid

        probs = np.reshape(np.array([0.88080, 0.5, 0.88080, 0.88080, 0.88080, 0.26894]), [6, 1])
        labels = np.reshape(np.array([1, 0, 1, 1, 1, 0]), [6, 1])

        assert np.array_equal(network.predict(X), labels) and np.allclose(network.predicted_probs, probs, atol=0.0001)

    def test_predict_bias(self):
        """Test predict method with bias term"""
        network = LogisticRegression(learning_rate=1, clip_max=1, max_attempts=100)

        X = np.array([[0, 1, 0, 1], [0, 0, 0, 0], [1, 1, 1, 1], [1, 1, 1, 1], [0, 0, 1, 1], [1, 0, 0, 0]])

        network.fitted_weights = np.array([-1, 1, 1, 1, -1])
        network.node_list = [5, 1]
        network.output_activation = sigmoid

        probs = np.reshape(np.array([0.73106, 0.26894, 0.73106, 0.73106, 0.73106, 0.11920]), [6, 1])
        labels = np.reshape(np.array([1, 0, 1, 1, 1, 0]), [6, 1])

        assert np.array_equal(network.predict(X), labels) and np.allclose(network.predicted_probs, probs, atol=0.0001)



"""Unit tests for algorithms/decorators"""

# Author: Kyle Nakamura
# License: BSD 3 clause

try:
    import ga_framework
except ImportError:
    import sys

    sys.path.append("..")

from mlrose_ky.decorators import short_name, get_short_name


def test_short_name_assignment():
    """Test that the short_name decorator assigns the correct short name to a function."""

    # noinspection PyMissingOrEmptyDocstring
    @short_name("SomeFunc")
    def some_function():
        return True

    assert hasattr(some_function, "__short_name__"), "The function should have a '__short_name__' attribute"
    assert some_function.__short_name__ == "SomeFunc", "The short name should be 'SomeFunc'"


def test_get_short_name_with_assigned_name():
    """Test retrieving the short name when it has been assigned."""

    # noinspection PyMissingOrEmptyDocstring
    @short_name("SomeFunc")
    def some_function():
        return True

    assert get_short_name(some_function) == "SomeFunc", "The short name should be 'SomeFunc'"


def test_get_short_name_without_assigned_name():
    """Test that the default function name is returned when no short name is assigned."""

    # noinspection PyMissingOrEmptyDocstring
    def some_function():
        return False

    assert get_short_name(some_function) == "some_function", "Should return the default function name 'some_function'"



"""Neural network activation functions."""

# Authors: Genevieve Hayes (modified by Andrew Rollings, Kyle Nakamura)
# Contributor : Ankit Grover
# License: BSD 3 clause
from mlrose_ky.decorators import short_name

import numpy as np

import warnings

warnings.filterwarnings("ignore")


@short_name("leaky_relu")
def leaky_relu(x, alpha=0.3, deriv=False):
    """Leaky ReLU activation function

    Parameters
    ----------
    x: np.ndarray
        Array containing input data.
    alpha: int , default : 0.3
        Alpha value to be set for applying small negative gradient
    deriv: bool, default: False
        Whether to return the function or its derivative.
        Set True for derivative.

    Returns
    -------
    fx: np.ndarray
        Value of activation function at x
    """
    fx = np.copy(x)
    fx = np.where(fx < 0, fx * alpha, fx)

    if deriv:
        fx[np.where(fx > 0)] = 1
        fx[np.where(fx < 0)] = alpha

    return fx



"""Neural network activation functions."""

# Authors: Genevieve Hayes (modified by Andrew Rollings, Kyle Nakamura)
# License: BSD 3 clause
from mlrose_ky.decorators import short_name

import numpy as np

import warnings

warnings.filterwarnings("ignore")


@short_name("identity")
def identity(x, deriv=False):
    """Linear activation function

    Parameters
    ----------
    x: np.ndarray
        Array containing input data.

    deriv: bool, default: False
        Whether to return the function or its derivative.
        Set True for derivative.

    Returns
    -------
    fx: np.ndarray
        Value of activation function at x
    """
    if not deriv:
        fx = x
    else:
        fx = np.ones(np.shape(x))

    return fx



"""Unit tests for gridsearch/"""

# Author: Kyle Nakamura
# License: BSD 3 clause

import pytest
import inspect
import numpy as np
from typing import Any
import sklearn.model_selection as skms
from sklearn.dummy import DummyClassifier
from sklearn.datasets import make_classification
from sklearn.model_selection import train_test_split

try:
    import mlrose_ky
except ImportError:
    import sys

    sys.path.append("..")
    import mlrose_ky

from mlrose_ky.gridsearch import GridSearchMixin

SEED = 12


@pytest.fixture
def sample_data():
    """Fixture for generating a simple dataset."""
    X, y = make_classification(random_state=SEED)
    return train_test_split(X, y, test_size=0.2, random_state=SEED)


@pytest.fixture
def dummy_classifier():
    """Fixture for creating a simple DummyClassifier."""
    return DummyClassifier(strategy="most_frequent", random_state=SEED)


@pytest.fixture
def grid_search_mixin():
    """Fixture for initializing a GridSearchMixin."""
    return GridSearchMixin()


@pytest.fixture
def grid_search_parameters():
    """Fixture for defining grid search parameters."""
    return {"strategy": ["most_frequent", "stratified"]}


class TestGridSearchMixin:

    def test_initialize_with_custom_scorer(self, grid_search_mixin):
        """Test initializing GridSearchMixin with a custom scoring method"""

        # noinspection PyMissingOrEmptyDocstring, PyShadowingNames
        def custom_scorer(y_true, y_pred):
            return np.mean(y_true == y_pred)

        grid_search_mixin.scorer_method = custom_scorer
        grid_search_mixin.params = inspect.signature(custom_scorer)  # Force update to match custom scorer

        assert grid_search_mixin.scorer_method == custom_scorer
        assert grid_search_mixin.params == inspect.signature(custom_scorer)

    @pytest.mark.filterwarnings("ignore::UserWarning")
    @pytest.mark.parametrize(
        "param_grid, expected_best_param",
        [({"strategy": ["most_frequent", "stratified"]}, "most_frequent"), ({"strategy": ["stratified"]}, "stratified")],
    )
    def test_perform_grid_search(self, grid_search_mixin, sample_data, dummy_classifier, param_grid, expected_best_param):
        """Should perform grid search with a classifier and return GridSearchCV object"""
        X_train, X_test, y_train, y_test = sample_data

        search_results = grid_search_mixin.perform_grid_search(
            classifier=dummy_classifier, x_train=X_train, y_train=y_train, cv=3, parameters=param_grid
        )

        assert isinstance(search_results, skms.GridSearchCV)
        assert hasattr(search_results, "best_params_")
        assert hasattr(search_results, "best_score_")
        assert hasattr(search_results, "cv_results_")
        assert search_results.best_params_["strategy"] == expected_best_param

    def test_handle_empty_input(self, grid_search_mixin):
        """Should handle empty input data gracefully"""
        with pytest.raises(ValueError):
            grid_search_mixin.perform_grid_search(
                classifier=DummyClassifier(), x_train=np.array([]), y_train=np.array([]), cv=3, parameters={"strategy": ["most_frequent"]}
            )

    def test_grid_search_score_intercept_raises_value_error(self, grid_search_mixin):
        """Should raise ValueError when scoring method raises TypeError"""

        # noinspection PyMissingOrEmptyDocstring, PyShadowingNames
        def faulty_scorer(_, __):
            raise TypeError("Intentional TypeError for testing")

        grid_search_mixin.scorer_method = faulty_scorer
        y_true = np.array([0, 1, 1, 0])
        y_pred = np.array([0, 0, 1, 1])

        with pytest.raises(ValueError, match="Error during scoring with method faulty_scorer"):
            grid_search_mixin._grid_search_score_intercept(y_pred=y_pred, y_true=y_true)

    def test_handle_multi_class_predictions_without_argmax(self, grid_search_mixin):
        """Test handling of multi-class predictions without applying argmax"""
        y_true = np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]])
        y_pred = np.array([[0.8, 0.1, 0.1], [0.2, 0.7, 0.1], [0.1, 0.2, 0.7]])

        # noinspection PyMissingOrEmptyDocstring, PyShadowingNames
        def custom_scorer(y_true, y_pred):
            return np.mean(np.argmax(y_true, axis=1) == np.argmax(y_pred, axis=1))

        grid_search_mixin.scorer_method = custom_scorer
        grid_search_mixin.params = inspect.signature(custom_scorer)
        score = grid_search_mixin._grid_search_score_intercept(y_pred=y_pred, y_true=y_true)

        assert score == 1.0

    def test_apply_argmax_to_multiclass_predictions(self, grid_search_mixin):
        """Should apply argmax to multi-class predictions when get_y_argmax is True"""
        grid_search_mixin.get_y_argmax = True

        y_true = np.array([[0, 1, 0], [1, 0, 0], [0, 0, 1]])
        y_pred = np.array([[0.1, 0.7, 0.2], [0.8, 0.1, 0.1], [0.2, 0.3, 0.5]])

        score = grid_search_mixin._grid_search_score_intercept(y_pred=y_pred, y_true=y_true)
        expected_score = grid_search_mixin.scorer_method(y_pred=y_pred.argmax(axis=1), y_true=y_true.argmax(axis=1))

        assert score == expected_score

    def test_make_scorer(self, grid_search_mixin):
        """Should create a custom scorer function using make_scorer method"""
        scorer: Any = grid_search_mixin.make_scorer()

        assert scorer._score_func == grid_search_mixin._grid_search_score_intercept
        assert callable(scorer)

    def test_score_with_additional_arguments(self, grid_search_mixin):
        """Should calculate score using the score method with additional arguments"""
        y_true = np.array([0, 1, 1, 0])
        y_pred = np.array([0, 0, 1, 1])

        # noinspection PyMissingOrEmptyDocstring, PyShadowingNames
        def custom_scorer(y_true, y_pred, weight=1.0):
            return np.mean(y_true == y_pred) * weight

        grid_search_mixin.scorer_method = custom_scorer
        grid_search_mixin.params = inspect.signature(custom_scorer)

        score = grid_search_mixin.score(y_pred=y_pred, y_true=y_true, weight=2.0)

        assert score == 1.0  # Expected score is 0.5 * 2.0

    def test_grid_search_score_intercept_with_additional_arguments(self, grid_search_mixin):
        """Should handle additional arguments in _grid_search_score_intercept method"""

        # noinspection PyMissingOrEmptyDocstring, PyShadowingNames
        def custom_scorer(y_true, y_pred, sample_weight=None):
            if sample_weight is not None:
                return np.average(y_true == y_pred, weights=sample_weight)
            return np.mean(y_true == y_pred)

        grid_search_mixin.scorer_method = custom_scorer
        y_true = np.array([0, 1, 1, 0])
        y_pred = np.array([0, 0, 1, 1])
        sample_weight = np.array([0.5, 0.5, 1.0, 1.0])

        score = grid_search_mixin._grid_search_score_intercept(y_pred=y_pred, y_true=y_true, sample_weight=sample_weight)
        expected_score = custom_scorer(y_true=y_true, y_pred=y_pred, sample_weight=sample_weight)

        assert score == expected_score

    @pytest.mark.filterwarnings("ignore::UserWarning")
    def test_perform_grid_search_with_verbose(self, grid_search_mixin, sample_data, dummy_classifier, grid_search_parameters):
        """Should perform grid search with verbose output enabled"""
        X_train, X_test, y_train, y_test = sample_data

        # Perform grid search with verbose output enabled
        search_results = grid_search_mixin.perform_grid_search(
            dummy_classifier, X_train, y_train, cv=3, parameters=grid_search_parameters, verbose=True
        )

        # Check that the result is a GridSearchCV object and has the expected attributes
        assert isinstance(search_results, skms.GridSearchCV)
        assert hasattr(search_results, "best_params_")
        assert hasattr(search_results, "best_score_")
        assert hasattr(search_results, "cv_results_")

    def test_perform_grid_search_with_multiple_jobs(self, grid_search_mixin, sample_data, dummy_classifier, grid_search_parameters):
        """Should perform grid search with n_jobs set to more than one"""
        X_train, X_test, y_train, y_test = sample_data

        # Perform grid search with n_jobs set to 2
        search_results = grid_search_mixin.perform_grid_search(
            dummy_classifier, X_train, y_train, cv=3, parameters=grid_search_parameters, n_jobs=2
        )

        # Check that the result is a GridSearchCV object and has the expected attributes
        assert isinstance(search_results, skms.GridSearchCV)
        assert hasattr(search_results, "best_params_")
        assert hasattr(search_results, "best_score_")
        assert hasattr(search_results, "cv_results_")



"""Classes for grid search functionality."""

# TODO: implement Bayesian Optimization for hyperparameter tuning, then rename this dir to 'hyperparameter_tuning'

# Authors: Genevieve Hayes (modified by Andrew Rollings, Kyle Nakamura)
# License: BSD 3 clause

from .grid_search_mixin import GridSearchMixin



"""
Provides decorators to assign and retrieve short names for functions and methods.

This file contains decorators that allow for the easy assignment of a 'short name'
to functions and methods, which can be particularly useful for assigning an algorithm's
common or shortened name, such as 'GA' for 'Genetic Algorithms'.

It also includes a function to retrieve the assigned short name of a function.
"""

# Authors: Genevieve Hayes (modified by Andrew Rollings, Kyle Nakamura)
# License: BSD 3 clause

from typing import Callable, Any


def short_name(name_expression: str) -> Callable:
    """
    Decorator to assign a short name to a function.

    Parameters
    ----------
    name_expression : str
        The short name to be assigned to the function.

    Returns
    -------
    Callable
        A decorator that assigns the provided short name to a function and returns the function.
    """

    def decorator(func: Callable) -> Callable:
        """Assign a short name to the given function."""
        func.__short_name__ = name_expression
        return func

    return decorator


def get_short_name(func: Any) -> str:
    """
    Retrieve the short name of a function, or its default name if a short name isn't assigned.

    Parameters
    ----------
    func : Any
        The function from which the short name is retrieved.

    Returns
    -------
    str
        The short name of the function, if assigned; otherwise, returns the full function name.
    """
    return getattr(func, "__short_name__", func.__name__)



"""Gene Swap Mutator for Genetic Algorithms (GA)."""

# Authors: Genevieve Hayes (modified by Andrew Rollings, Kyle Nakamura)
# License: BSD 3 clause

import numpy as np
from typing import Any

from mlrose_ky.algorithms.mutators._mutator_base import _MutatorBase


class GeneSwapMutator(_MutatorBase):
    """
    A mutator class that implements the 'Gene Swap' mutation strategy in genetic algorithms.

    This mutator randomly selects two genes in a chromosome and swaps their positions,
    thereby altering the genetic makeup of the offspring without introducing external genetic material.

    Attributes
    ----------
    optimization_problem : Any
        The optimization problem instance associated with the mutation operations.

    Parameters
    ----------
    optimization_problem : OptimizationProblem
        An instance of an optimization problem that the mutator will operate on.
    """

    def __init__(self, optimization_problem: Any) -> None:
        super().__init__(optimization_problem)

    def mutate(self, child: np.ndarray, mutation_probability: float) -> np.ndarray:
        """
        Perform a gene swap mutation on the given chromosome if the mutation probability condition is met.

        Two genes are randomly selected and their positions are swapped. This mutation method helps maintain
        genetic diversity within the population.

        Parameters
        ----------
        child : np.ndarray
            The chromosome of a child individual to be mutated.
        mutation_probability : float
            The probability of the child undergoing a mutation.

        Returns
        -------
        np.ndarray
            The chromosome after mutation.
        """
        if np.random.rand() < mutation_probability:
            index_one, index_two = np.random.choice(len(child), size=2, replace=False)
            child[index_one], child[index_two] = child[index_two], child[index_one]

        return child



"""Classes for defining fitness functions (i.e., optimization problems) for optimization algorithms."""

from .continuous_peaks import ContinuousPeaks

from .flip_flop import FlipFlop

from .four_peaks import FourPeaks

from .knapsack import Knapsack

from .max_k_color import MaxKColor

from .one_max import OneMax

from .queens import Queens

from .six_peaks import SixPeaks

from .travelling_salesperson import TravellingSalesperson

from .custom_fitness import CustomFitness



"""Class defining a Max-K Color optimization problem generator."""

# Authors: Genevieve Hayes (modified by Andrew Rollings, Kyle Nakamura)
# License: BSD 3 clause

import numpy as np
import networkx as nx

from mlrose_ky.opt_probs import MaxKColorOpt


class MaxKColorGenerator:
    """A class to generate Max-K Color optimization problems."""

    @staticmethod
    def generate(
        seed: int, number_of_nodes: int = 20, max_connections_per_node: int = 4, max_colors: int | None = None, maximize: bool = False
    ) -> MaxKColorOpt:
        """
        Generate a Max-K Color optimization problem instance.

        Parameters
        ----------
        seed : int
            Seed for the random number generator.
        number_of_nodes : int, optional, default=20
            The number of nodes in the graph.
        max_connections_per_node : int, optional, default=4
            The maximum number of connections (edges) per node.
        max_colors : int or None, optional, default=None
            The maximum number of colors available.
        maximize : bool, optional, default=False
            Whether the optimization problem should be maximized.

        Returns
        -------
        MaxKColorOpt
            An instance of MaxKColorOpt configured with the specified parameters.

        Raises
        ------
        ValueError
            If any parameter is not of the expected type or value.

        Examples
        --------
        >>> import mlrose_ky
        >>> test_edges = [(0, 1), (0, 2), (0, 4), (1, 3), (2, 0), (2, 3), (3, 4)]
        >>> fitness = mlrose_ky.MaxKColor(test_edges)
        >>> state = np.array([0, 1, 0, 1, 1])
        >>> fitness.evaluate(state)
        3.0
        """
        if not isinstance(seed, int):
            raise ValueError(f"Seed must be an integer. Got {seed}")
        if not isinstance(number_of_nodes, int) or number_of_nodes <= 0:
            raise ValueError(f"Number of nodes must be a positive integer. Got {number_of_nodes}")
        if not isinstance(max_connections_per_node, int) or max_connections_per_node <= 0:
            raise ValueError(f"Max connections per node must be a positive integer. Got {max_connections_per_node}")
        if max_colors is not None and (not isinstance(max_colors, int) or max_colors <= 0):
            raise ValueError(f"Max colors must be a positive integer or None. Got {max_colors}")
        if not isinstance(maximize, bool):
            raise ValueError(f"Maximize must be a boolean. Got {maximize}")

        # Handle single node case
        if number_of_nodes == 1:
            return MaxKColorOpt(edges=[], length=1, maximize=maximize, max_colors=max_colors)

        np.random.seed(seed)

        # Generate random connection counts for each node
        node_connection_counts = 1 + np.random.choice(max_connections_per_node, size=number_of_nodes)

        node_connections = {}
        nodes = range(number_of_nodes)
        for node in nodes:
            valid_other_nodes = [
                other for other in nodes if (other != node and (other not in node_connections or node not in node_connections[other]))
            ]
            count = min(node_connection_counts[node], len(valid_other_nodes))
            connected_nodes = sorted(np.random.choice(valid_other_nodes, count, replace=False))
            node_connections[node] = [(node, other) for other in connected_nodes]

        # Ensure graph connectivity
        graph = nx.Graph()
        graph.add_edges_from([edge for edges in node_connections.values() for edge in edges])

        for node in nodes:
            unreachable = [
                (node, other) if node < other else (other, node) for other in nodes if other not in nx.bfs_tree(graph, node).nodes()
            ]
            for start, end in unreachable:
                graph.add_edge(start, end)
                remaining_unreachable = len(
                    [(node, other) if node < other else (other, node) for other in nodes if other not in nx.bfs_tree(graph, node).nodes()]
                )
                if remaining_unreachable == 0:
                    break

        problem = MaxKColorOpt(
            edges=list(graph.edges()), length=number_of_nodes, maximize=maximize, max_colors=max_colors, source_graph=graph
        )

        return problem



"""Class defining the One Max fitness function for use with optimization algorithms."""

# Authors: Genevieve Hayes (modified by Andrew Rollings, Kyle Nakamura)
# License: BSD 3 clause

import numpy as np


class OneMax:
    """Fitness function for One Max optimization problem.

    Evaluates the fitness of an n-dimensional state vector `x` as:

    .. math::

        Fitness(x) = \\sum_{i = 0}^{n-1}x_{i}

    Examples
    -------
    >>> fitness = OneMax()
    >>> state = np.array([0, 1, 0, 1, 1, 1, 1])
    >>> fitness.evaluate(state)
    5.0

    Note
    -----
    The One Max fitness function is suitable for use in either discrete or continuous-state optimization problems.
    """

    def __init__(self) -> None:
        self.problem_type: str = "either"

    @staticmethod
    def evaluate(state_vector: np.ndarray) -> float:
        """Evaluate the fitness of a state vector.

        Parameters
        ----------
        state_vector: np.ndarray
            State array for evaluation.

        Returns
        -------
        float
            Value of fitness function.

        Raises
        ------
        TypeError
            If `state_vector` is not an instance of `np.ndarray`.
        """
        if not isinstance(state_vector, np.ndarray):
            raise TypeError(f"Expected state_vector to be np.ndarray, got {type(state_vector).__name__} instead.")

        return float(np.sum(state_vector))

    def get_problem_type(self) -> str:
        """Return the problem type.

        Returns
        -------
        str
            Specifies problem type as 'either'.
        """
        return self.problem_type



"""Uniform Crossover implementation for Genetic Algorithms (GA).

This module defines a uniform crossover operation used in genetic algorithms,
where each gene in the offspring is chosen randomly from one of the corresponding
genes of the parents.
"""

# Authors: Genevieve Hayes (modified by Andrew Rollings, Kyle Nakamura)
# License: BSD 3 Clause

import numpy as np
from typing import Any, Sequence

from mlrose_ky.algorithms.crossovers._crossover_base import _CrossoverBase


class UniformCrossover(_CrossoverBase):
    """
    Uniform crossover for genetic algorithms.

    This class implements a uniform crossover operation where each gene of the offspring
    is independently chosen from one of the two parents with equal probability. This
    method is often used when no prior knowledge about the problem structure is known.

    Inherits from:
    _CrossoverBase : Abstract base class for crossover operations.
    """

    def __init__(self, optimization_problem: Any) -> None:
        """
        Initialize the UniformCrossover with the given optimization problem.

        Parameters
        ----------
        optimization_problem : Any
            An instance of the optimization problem related to the genetic algorithm.
        """
        super().__init__(optimization_problem)

    def mate(self, parent1: Sequence[float], parent2: Sequence[float]) -> np.ndarray:
        """
        Perform the uniform crossover between two parent sequences to produce offspring.

        Each gene of the offspring is randomly chosen from one of the two parents'
        corresponding genes with a 50% chance for each.

        Parameters
        ----------
        parent1 : Sequence[float]
            The first parent chromosome sequence.
        parent2 : Sequence[float]
            The second parent chromosome sequence.

        Returns
        -------
        np.ndarray
            The offspring chromosome resulting from the crossover.
        """
        gene_selector = np.random.randint(2, size=self.chromosome_length)
        stacked_parents = np.vstack((parent1, parent2))
        offspring = stacked_parents[gene_selector, np.arange(self.chromosome_length)]
        return offspring



"""Unit tests for generators/"""

# Author: Kyle Nakamura
# License: BSD 3 clause

from unittest.mock import patch

import matplotlib.pyplot as plt
import pytest
from sklearn.linear_model import LogisticRegression

try:
    import mlrose_ky
except ImportError:
    import sys

    sys.path.append("..")
    import mlrose_ky

from mlrose_ky.samples import SyntheticDataGenerator, plot_synthetic_dataset


# noinspection PyMissingOrEmptyDocstring
@pytest.fixture
def generator():
    return SyntheticDataGenerator(seed=42)


class TestSyntheticDataGenerator:
    def test_get_synthetic_features_and_classes_default(self, generator):
        """Test default synthetic features and classes."""
        features, classes = generator.get_synthetic_features_and_classes()

        assert features == ["(1) A", "(2) B"]
        assert classes == ["RED", "BLUE"]

    def test_get_synthetic_features_and_classes_with_redundant(self, generator):
        """Test synthetic features and classes with redundant column."""
        features, classes = generator.get_synthetic_features_and_classes(with_redundant_column=True)

        assert features == ["(1) A", "(2) B", "(3) R"]
        assert classes == ["RED", "BLUE"]

    def test_get_synthetic_data(self, generator):
        """Test getting synthetic data."""
        data, features, classes, output_directory = generator.get_synthetic_data()

        assert data.shape == (400, 3)
        assert features == ["(1) A", "(2) B"]
        assert classes == ["RED", "BLUE"]
        assert output_directory is None

    def test_setup_synthetic_data_test_train(self, generator):
        """Test setting up synthetic data for train and test."""
        data, _, _, _ = generator.get_synthetic_data()
        x, y, x_train, x_test, y_train, y_test = generator.setup_synthetic_data_test_train(data)

        assert x_train.shape == (280, 2)
        assert x_test.shape == (120, 2)
        assert y_train.shape == (280,)
        assert y_test.shape == (120,)

    @patch('matplotlib.pyplot.show')  # Mocking plt.show
    def test_plot_synthetic_dataset(self, mock_show, generator):
        """Test plotting synthetic dataset."""
        data, _, _, _ = generator.get_synthetic_data()
        x, y, x_train, x_test, y_train, y_test = generator.setup_synthetic_data_test_train(data)
        classifier = LogisticRegression().fit(x_train, y_train)
        plt.figure()
        plot_synthetic_dataset(x_train, x_test, y_train, y_test, classifier=classifier)
        plt.close()

        assert mock_show.called



import mlrose_ky.neural.activation as act
import sklearn.metrics as skmt

from sklearn.neural_network import MLPClassifier
from sklearn.base import BaseEstimator
from mlrose_ky.decorators import short_name
from mlrose_ky.runners._nn_runner_base import _NNRunnerBase


@short_name("skmlp")
class SKMLPRunner(_NNRunnerBase):
    class _MLPClassifier(BaseEstimator):
        def __init__(self, runner, **kwargs):
            self.runner = runner
            self.mlp = MLPClassifier(**kwargs)
            self.state_callback = self.runner.save_state
            self.fit_started_ = False
            self.user_info_ = None
            self.kwargs_ = kwargs

            self.loss_ = 1.0
            self.state_ = None
            self.curve_ = []

            # need to intercept the classifier so we can track statistics.
            if runner.generate_curves:
                if hasattr(self.mlp, "_update_no_improvement_count"):
                    # noinspection PyProtectedMember
                    self._mlp_update_no_improvement_count = self.mlp._update_no_improvement_count
                    self.mlp._update_no_improvement_count = self._update_no_improvement_count_intercept
                if hasattr(self.mlp, "_loss_grad_lbfgs"):
                    # noinspection PyProtectedMember
                    self._mlp_loss_grad_lbfgs = self.mlp._loss_grad_lbfgs
                    self.mlp._loss_grad_lbfgs = self._loss_grad_lbfgs_intercept

        def __getattr__(self, item, default=None):
            if "mlp" in self.__dict__ and hasattr(self.__dict__["mlp"], item):
                return self.__dict__["mlp"].__getattr__(item, default)
            return self.__dict__[item] if item in self.__dict__ else default

        def __setattr__(self, item, value):
            if "mlp" in self.__dict__ and hasattr(self.__dict__["mlp"], item):
                self.__dict__["mlp"].__setattr__(item, value)
            self.__dict__[item] = value

        def get_params(self, deep=True):
            out = super().get_params()
            out.update(self.mlp.get_params())
            # exclude any that end with an underscore
            out = {k: v for (k, v) in out.items() if not k[-1] == "_"}
            return out

        def fit(self, x_train, y_train=None):
            self.fit_started_ = True
            self.runner.start_run_timing()
            # make initial callback
            self._invoke_runner_callback()
            return self.mlp.fit(x_train, y_train)

        def predict(self, x_test):
            return self.mlp.predict(x_test)

        def _update_no_improvement_count_intercept(self, early_stopping, x_val, y_val):
            ret = self._mlp_update_no_improvement_count(early_stopping, x_val, y_val)
            self._state = self.mlp.coefs_ if hasattr(self.mlp, "coefs_") else []
            self.loss_ = self.mlp.loss_ if hasattr(self.mlp, "loss_") else 0
            if hasattr(self.mlp, "loss_curve_"):
                self.curve_ = [(_loss_val, None) for _loss_val in self.mlp.loss_curve_]
            else:
                self.curve_.append((self.loss_, None))
            self._invoke_runner_callback()
            return ret

        def _loss_grad_lbfgs_intercept(self, packed_coef_inter, x, y, activations, deltas, coef_grads, intercept_grads):
            f, g = self._mlp_loss_grad_lbfgs(packed_coef_inter, x, y, activations, deltas, coef_grads, intercept_grads)
            self.loss_ = f
            self.state_ = g
            self.curve_.append((self.loss_, None))
            self._invoke_runner_callback()
            return f, g

        def _invoke_runner_callback(self):
            iterations = self.mlp.n_iter_ if hasattr(self.mlp, "n_iter_") else 0
            # noinspection PyProtectedMember
            no_improvement_count = self.mlp._no_improvement_count if hasattr(self.mlp, "_no_improvement_count") else 0

            done = self.mlp.early_stopping and (no_improvement_count > self.mlp.n_iter_no_change) or iterations == self.mlp.max_iter

            # check for early abort.
            if self.runner.has_aborted():
                return self
            if self.user_info_ is None:
                self.user_info_ = [(k, self.__dict__[k]) for k in self.kwargs_.keys() if hasattr(self, k)]
                for k, v in self.user_info_:
                    self.runner.log_current_argument(k, v)

            return self.state_callback(
                iteration=iterations,
                state=self.state_,
                fitness=self.loss_,
                user_data=self.user_info_,
                attempt=no_improvement_count,
                done=done,
                curve=self.curve_,
            )

    def __init__(
        self,
        x_train,
        y_train,
        x_test,
        y_test,
        experiment_name,
        seed,
        iteration_list,
        grid_search_parameters,
        grid_search_scorer_method=skmt.balanced_accuracy_score,
        early_stopping=True,
        max_attempts=500,
        n_jobs=1,
        cv=5,
        override_ctrl_c_handler=True,
        generate_curves=True,
        output_directory=None,
        replay=False,
        **kwargs,
    ):

        # take a copy of the grid search parameters
        grid_search_parameters = {**grid_search_parameters}

        # hack for compatibility purposes
        if "max_iters" in grid_search_parameters:
            grid_search_parameters["max_iter"] = grid_search_parameters.pop("max_iters")

        if "max_attempts" in grid_search_parameters:
            grid_search_parameters["n_iter_no_change"] = grid_search_parameters.pop("max_attempts")

        super().__init__(
            x_train=x_train,
            y_train=y_train,
            x_test=x_test,
            y_test=y_test,
            experiment_name=experiment_name,
            seed=seed,
            iteration_list=iteration_list,
            grid_search_parameters=grid_search_parameters,
            grid_search_scorer_method=grid_search_scorer_method,
            generate_curves=generate_curves,
            output_directory=output_directory,
            override_ctrl_c_handler=override_ctrl_c_handler,
            replay=replay,
            n_jobs=n_jobs,
            cv=cv,
        )

        # build the classifier
        self.classifier = self._MLPClassifier(
            runner=self,
            shuffle=True,
            random_state=seed,
            verbose=False,
            warm_start=False,
            early_stopping=early_stopping,
            n_iter_no_change=max_attempts,
            **kwargs,
        )

        self.classifier.runner = self

    @staticmethod
    def build_grid_search_parameters(grid_search_parameters, **kwargs):
        # extract nn parameters
        all_grid_search_parameters = _NNRunnerBase.build_grid_search_parameters(grid_search_parameters, **kwargs)
        # make sure activation set is the right type.
        if "activation" in all_grid_search_parameters:
            activation_set = list(all_grid_search_parameters["activation"])
            for i in range(len(activation_set)):
                a = activation_set[i]
                if a == act.relu:
                    activation_set[i] = "relu"
                elif a == act.sigmoid:
                    activation_set[i] = "logistic"
                elif a == act.tanh:
                    activation_set[i] = "tanh"
                elif a == act.identity:
                    activation_set[i] = "identity"
                elif a == act.softmax:
                    activation_set[i] = "softmax"
            all_grid_search_parameters["activation"] = activation_set
        return all_grid_search_parameters



"""Class for defining an arithmetic decay schedule for Simulated Annealing (SA)."""

# Authors: Genevieve Hayes (modified by Andrew Rollings, Kyle Nakamura)
# License: BSD 3 clause


class ArithmeticDecay:
    """
    Schedule for arithmetically decaying the temperature parameter T in a
    simulated annealing process, calculated using the formula:

    .. math::

        T(t) = \\max(T_{0} - rt, T_{min})

    where :math:`T_{0}` is the initial temperature at time `t = 0`, `r` is the decay rate, and :math:`T_{min}`
    is the minimum temperature.

    Parameters
    ----------
    initial_temperature : float
        Initial value of the temperature parameter T. Must be greater than 0.
    decay_rate : float
        Temperature decay parameter. Must be a positive value and less than or equal to 1.
    minimum_temperature : float
        Minimum allowable value of the temperature parameter. Must be positive and less than `initial_temperature`.

    Attributes
    ----------
    initial_temperature : float
        Stores the initial temperature.
    decay_rate : float
        Stores the rate of temperature decay.
    minimum_temperature : float
        Stores the minimum temperature.

    Examples
    --------
    >>> schedule = ArithmeticDecay(initial_temperature=10, decay_rate=0.95, minimum_temperature=1)
    >>> schedule.evaluate(5)
    5.25
    """

    def __init__(self, initial_temperature: float = 1.0, decay_rate: float = 0.0001, minimum_temperature: float = 0.001) -> None:
        self.initial_temperature: float = initial_temperature
        self.decay_rate: float = decay_rate
        self.minimum_temperature: float = minimum_temperature

        if self.initial_temperature <= 0:
            raise ValueError("Initial temperature must be greater than 0.")
        if not (0 < self.decay_rate <= 1):
            raise ValueError("Decay rate must be greater than 0 and less than or equal to 1.")
        if not (0 < self.minimum_temperature < self.initial_temperature):
            raise ValueError("Minimum temperature must be greater than 0 and less than initial temperature.")

    def __str__(self) -> str:
        return (
            f"ArithmeticDecay(initial_temperature={self.initial_temperature}, "
            f"decay_rate={self.decay_rate}, "
            f"minimum_temperature={self.minimum_temperature})"
        )

    def __repr__(self) -> str:
        return self.__str__()

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, ArithmeticDecay):
            return False
        return (
            self.initial_temperature == other.initial_temperature
            and self.decay_rate == other.decay_rate
            and self.minimum_temperature == other.minimum_temperature
        )

    def evaluate(self, time: int) -> float:
        """
        Calculate and return the temperature parameter at the given time.

        Parameters
        ----------
        time : int
            The time at which to evaluate the temperature parameter.

        Returns
        -------
        float
            The temperature parameter at the given time, respecting the minimum temperature.
        """
        temperature = max(self.initial_temperature - (self.decay_rate * time), self.minimum_temperature)
        return temperature

    def get_info(self, time: int = None, prefix: str = "") -> dict:
        """
        Generate a dictionary containing the decay schedule's settings and optionally its current value.

        Parameters
        ----------
        time : int, optional
            The time at which to evaluate the current temperature value.
            If provided, the current value is included in the returned dictionary.
        prefix : str, optional
            A prefix to prepend to each key in the dictionary, useful for integrating this info into larger structured data.

        Returns
        -------
        dict
            A dictionary with keys reflecting the decay schedule's parameters and optionally the current temperature.
        """
        info_prefix = f"{prefix}schedule_" if prefix else "schedule_"

        info = {
            f"{info_prefix}type": "arithmetic",
            f"{info_prefix}initial_temperature": self.initial_temperature,
            f"{info_prefix}decay_rate": self.decay_rate,
            f"{info_prefix}minimum_temperature": self.minimum_temperature,
        }

        if time is not None:
            info[f"{info_prefix}current_value"] = self.evaluate(time)

        return info



"""Unit tests for runners/"""

# Author: Kyle Nakamura
# License: BSD 3 clause

# import numpy as np

try:
    import mlrose_ky
except ImportError:
    import sys

    sys.path.append("..")
    import mlrose_ky



"""Classes for defining optimization problem objects."""

# Authors: Genevieve Hayes (modified by Andrew Rollings, Kyle Nakamura)
# License: BSD 3 clause

import numpy as np

from mlrose_ky.opt_probs.opt_prob import OptProb


class ContinuousOpt(OptProb):
    """Class for defining continuous-state optimisation problems.

    Parameters
    ----------
    length: int
        Number of elements in state vector.

    fitness_fn: fitness function object
        Object to implement fitness function for optimization.

    maximize: bool, default: True
        Whether to maximize the fitness function.
        Set :code:`False` for minimization problem.

    min_val: float, default: 0
        Minimum value that each element of the state vector can take.

    max_val: float, default: 1
        Maximum value that each element of the state vector can take.

    step: float, default: 0.1
        Step size used in determining neighbors of current state.
    """

    def __init__(self, length, fitness_fn, maximize=True, min_val=0, max_val=1, step=0.1):

        OptProb.__init__(self, length, fitness_fn, maximize=maximize)

        if (self.fitness_fn.get_problem_type() != "continuous") and (self.fitness_fn.get_problem_type() != "either"):
            raise Exception(
                "fitness_fn must have problem type 'continuous'"
                + """ or 'either'. Define problem as"""
                + """ DiscreteOpt problem or use alternative"""
                + """ fitness function."""
            )

        if max_val <= min_val:
            raise Exception("""max_val must be greater than min_val.""")

        if step <= 0:
            raise Exception("""step size must be positive.""")

        if (max_val - min_val) < step:
            raise Exception("""step size must be less than""" + """ (max_val - min_val).""")

        self.min_val = min_val
        self.max_val = max_val
        self.step = step
        self.prob_type = "continuous"

    def calculate_updates(self):
        """Calculate gradient descent updates.

        Returns
        -------
        updates: list
            List of back propagation weight updates.
        """
        updates = self.fitness_fn.calculate_updates()

        return updates

    def find_neighbors(self):
        """Find all neighbors of the current state."""

        self.neighbors = []

        for i in range(self.length):
            for j in [-1, 1]:
                neighbor = np.copy(self.state)
                neighbor[i] += j * self.step

                if neighbor[i] > self.max_val:
                    neighbor[i] = self.max_val

                elif neighbor[i] < self.min_val:
                    neighbor[i] = self.min_val

                if not np.array_equal(np.array(neighbor), self.state):
                    self.neighbors.append(neighbor)

    def get_problem_type(self):
        """Return the problem type.

        Returns
        -------
        self.prob_type: str
            Returns problem type.
        """
        return self.prob_type

    def random(self):
        """Return a random state vector.

        Returns
        -------
        state: np.ndarray
            Randomly generated state vector.
        """
        state = np.random.uniform(self.min_val, self.max_val, self.length)

        return state

    def random_neighbor(self):
        """Return random neighbor of current state vector.

        Returns
        -------
        neighbor: np.ndarray
            State vector of random neighbor.
        """
        while True:
            neighbor = np.copy(self.state)
            i = np.random.randint(0, self.length)

            neighbor[i] += self.step * np.random.choice([-1, 1])

            if neighbor[i] > self.max_val:
                neighbor[i] = self.max_val

            elif neighbor[i] < self.min_val:
                neighbor[i] = self.min_val

            if not np.array_equal(np.array(neighbor), self.state):
                break

        return neighbor

    def random_pop(self, pop_size):
        """Create a population of random state vectors.

        Parameters
        ----------
        pop_size: int
            Size of population to be created.
        """
        if pop_size <= 0:
            raise Exception("""pop_size must be a positive integer.""")
        elif not isinstance(pop_size, int):
            if pop_size.is_integer():
                pop_size = int(pop_size)
            else:
                raise Exception("""pop_size must be a positive integer.""")

        population = []
        pop_fitness = []

        for _ in range(pop_size):
            state = self.random()
            fitness = self.eval_fitness(state)

            population.append(state)
            pop_fitness.append(fitness)

        self.population = np.array(population)
        self.pop_fitness = np.array(pop_fitness)

    def reproduce(self, parent_1, parent_2, mutation_prob=0.1):
        """Create child state vector from two parent state vectors.

        Parameters
        ----------
        parent_1: np.ndarray
            State vector for parent 1.

        parent_2: np.ndarray
            State vector for parent 2.

        mutation_prob: float
            Probability of a mutation at each state vector element during
            reproduction.

        Returns
        -------
        child: np.ndarray
            Child state vector produced from parents 1 and 2.
        """
        if len(parent_1) != self.length or len(parent_2) != self.length:
            raise Exception("""Lengths of parents must match problem length""")

        if (mutation_prob < 0) or (mutation_prob > 1):
            raise Exception("""mutation_prob must be between 0 and 1.""")

        # Reproduce parents
        if self.length > 1:
            _n = np.random.randint(self.length - 1)
            child = np.array([0.0] * self.length)
            child[0 : _n + 1] = parent_1[0 : _n + 1]
            child[_n + 1 :] = parent_2[_n + 1 :]
        elif np.random.randint(2) == 0:
            child = np.copy(parent_1)
        else:
            child = np.copy(parent_2)

        # Mutate child
        rand = np.random.uniform(size=self.length)
        mutate = np.where(rand < mutation_prob)[0]

        for i in mutate:
            child[i] = np.random.uniform(self.min_val, self.max_val)

        return child

    def reset(self):
        """Set the current state vector to a random value and get its fitness."""
        self.state = self.random()
        self.fitness_evaluations = 0
        self.fitness = self.eval_fitness(self.state)

    def update_state(self, updates):
        """Update current state given a vector of updates.

        Parameters
        ----------
        updates: np.ndarray
            Update array.

        Returns
        -------
        updated_state: np.ndarray
            Current state adjusted for updates.
        """
        if len(updates) != self.length:
            raise Exception("""Length of updates must match problem length""")

        updated_state = self.state + updates

        updated_state[updated_state > self.max_val] = self.max_val
        updated_state[updated_state < self.min_val] = self.min_val

        return updated_state



import mlrose_ky
from mlrose_ky.decorators import short_name
from mlrose_ky.runners._runner_base import _RunnerBase
import numpy as np

"""
Example usage:

    experiment_name = 'example_experiment'
    problem = TSPGenerator.generate(seed=SEED, number_of_cities=22)
    
    # note that you can also initialize a temperature_list this way
    # temperature_list = [mlrose_ky.GeometricDecay(init_temp=t, decay=d) for (t, d) in [(1, 0.99), (1e2, 0.999)]]
    # if you use this form, the decay_list parameter is ignored.

    sa = SARunner(problem=problem,
                  experiment_name=experiment_name,
                  output_directory=OUTPUT_DIRECTORY,
                  seed=SEED,
                  iteration_list=2 ** np.arange(14),
                  max_attempts=5000,
                  temperature_list=[1, 10, 50, 100, 250, 500, 1000, 2500, 5000, 10000],
                  decay_list=[mlrose_ky.GeometricDecay])

    # the two data frames will contain the results
    df_run_stats, df_run_curves = sa.run()                  
"""


@short_name("sa")
class SARunner(_RunnerBase):

    def __init__(
        self,
        problem,
        experiment_name,
        seed,
        iteration_list,
        temperature_list,
        decay_list=None,
        max_attempts=500,
        generate_curves=True,
        **kwargs,
    ):
        super().__init__(
            problem=problem,
            experiment_name=experiment_name,
            seed=seed,
            iteration_list=iteration_list,
            max_attempts=max_attempts,
            generate_curves=generate_curves,
            **kwargs,
        )
        self.use_raw_temp = True
        self.temperature_list = temperature_list
        if all([np.isscalar(x) for x in temperature_list]):
            if decay_list is None:
                decay_list = [mlrose_ky.GeometricDecay]
            self.decay_list = decay_list
            self.use_raw_temp = False

    def run(self):
        temperatures = (
            self.temperature_list
            if self.use_raw_temp
            else [d(initial_temperature=t) for t in self.temperature_list for d in self.decay_list]
        )
        return super().run_experiment_(algorithm=mlrose_ky.simulated_annealing, schedule=("Temperature", temperatures))



"""Classes for defining neural network weight optimization problems."""

# Authors: Genevieve Hayes (modified by Andrew Rollings, Kyle Nakamura)
# License: BSD 3 clause

from .weights import flatten_weights, unflatten_weights, gradient_descent_original



"""Class defining a Traveling Salesman Problem (TSP) optimization problem generator."""

# Authors: Genevieve Hayes (modified by Andrew Rollings, Kyle Nakamura)
# License: BSD 3 clause

import numpy as np
import networkx as nx
import itertools as it
from collections import defaultdict

from mlrose_ky import TSPOpt


class TSPGenerator:
    """A class to generate Traveling Salesman Problem (TSP) optimization problems."""

    @staticmethod
    def generate(seed: int, number_of_cities: int, area_width: int = 250, area_height: int = 250) -> TSPOpt:
        """
        Generate a TSP optimization problem instance.

        Parameters
        ----------
        seed : int
            Seed for the random number generator.
        number_of_cities : int
            The number of cities (nodes) in the TSP.
        area_width : int, optional, default=250
            The width of the area in which cities are placed.
        area_height : int, optional, default=250
            The height of the area in which cities are placed.

        Returns
        -------
        TSPOpt
            An instance of TSPOpt configured with the specified parameters.

        Raises
        ------
        ValueError
            If any of the parameters are not of the expected type or value.
        """
        if not isinstance(seed, int):
            raise ValueError(f"Seed must be an integer. Got {seed}")
        if not isinstance(number_of_cities, int) or number_of_cities <= 0:
            raise ValueError(f"Number of cities must be a positive integer. Got {number_of_cities}")
        if not isinstance(area_width, int) or area_width <= 0:
            raise ValueError(f"Area width must be a positive integer. Got {area_width}")
        if not isinstance(area_height, int) or area_height <= 0:
            raise ValueError(f"Area height must be a positive integer. Got {area_height}")

        # Generate random coordinates for cities
        np.random.seed(seed)
        x_coords = np.random.randint(area_width, size=number_of_cities)
        y_coords = np.random.randint(area_height, size=number_of_cities)

        coords = list(tuple(zip(x_coords, y_coords)))
        duplicates = TSPGenerator.list_duplicates_(coords)

        # Ensure no duplicate coordinates
        while len(duplicates) > 0:
            for d in duplicates:
                x_coords = np.random.randint(area_width, size=len(d))
                y_coords = np.random.randint(area_height, size=len(d))
                for i in range(len(d)):
                    coords[d[i]] = (x_coords[i], y_coords[i])
            duplicates = TSPGenerator.list_duplicates_(coords)

        # Calculate distances between all pairs of cities
        distances = TSPGenerator.get_distances(coords, truncate=False)

        # Create a graph with the calculated distances
        graph = nx.Graph()
        for a, b, distance in distances:
            graph.add_edge(a, b, length=int(round(distance)))

        return TSPOpt(coords=coords, distances=distances, source_graph=graph)

    @staticmethod
    def get_distances(coords: list[tuple], truncate: bool = True) -> list[tuple]:
        """
        Calculate the distances between all pairs of coordinates.

        Parameters
        ----------
        coords : list[tuple]
            A list of (x, y) coordinates representing the cities.
        truncate : bool, optional, default=True
            If True, truncate the distances to integers.

        Returns
        -------
        distances : list[tuple]
            A list of tuples representing the distance between each pair of cities.
        """
        # Calculate Euclidean distances between all pairs of coordinates
        distances = [
            (c1, c2, np.linalg.norm(np.subtract(coords[c1], coords[c2])))
            for c1, c2 in it.product(range(len(coords)), range(len(coords)))
            if c1 != c2 and c2 > c1
        ]
        if truncate:
            distances = [(c1, c2, int(d)) for c1, c2, d in distances]

        return distances

    @staticmethod
    def list_duplicates_(seq: list[tuple]) -> list[list[int]]:
        """
        Identify duplicate entries in the sequence.

        Parameters
        ----------
        seq : list[tuple]
            A list of (x, y) coordinates.

        Returns
        -------
        list of list of int
            A list containing lists of indices for duplicate coordinates.
        """
        # Create a tally of all coordinates
        tally = defaultdict(list)

        for i, item in enumerate(seq):
            tally[item].append(i)

        # Return indices of duplicate coordinates
        return [indices[1:] for _, indices in tally.items() if len(indices) > 1]



"""Class defining a One Max optimization problem generator."""

# Authors: Genevieve Hayes (modified by Andrew Rollings, Kyle Nakamura)
# License: BSD 3 clause

import numpy as np

from mlrose_ky import DiscreteOpt, OneMax


class OneMaxGenerator:
    """A class to generate One Max optimization problems."""

    @staticmethod
    def generate(seed: int, size: int = 20) -> DiscreteOpt:
        """
        Generate a One Max optimization problem instance.

        Parameters
        ----------
        seed : int
            Seed for the random number generator.
        size : int, optional, default=20
            The size of the optimization problem (number of bits).

        Returns
        -------
        DiscreteOpt
            An instance of DiscreteOpt configured for the One Max problem.

        Raises
        ------
        ValueError
            If `size` is not a positive integer.
        """
        if not isinstance(seed, int):
            raise ValueError(f"Seed must be an integer. Got {seed}")
        if not isinstance(size, int) or size <= 0:
            raise ValueError(f"Size must be a positive integer. Got {size}")

        np.random.seed(seed)

        fitness = OneMax()
        problem = DiscreteOpt(length=size, fitness_fn=fitness)

        return problem



"""Classes for defining optimization problem objects."""

# Authors: Genevieve Hayes (modified by Andrew Rollings, Kyle Nakamura)
# License: BSD 3 clause

import numpy as np
from sklearn.metrics import mutual_info_score
from scipy.sparse import csr_matrix
from scipy.sparse.csgraph import minimum_spanning_tree, depth_first_tree

from mlrose_ky.algorithms.crossovers import UniformCrossover
from mlrose_ky.algorithms.mutators import GeneSwapMutator
from mlrose_ky.opt_probs.opt_prob import OptProb


class DiscreteOpt(OptProb):
    """Class for defining discrete-state optimization problems.

    Parameters
    ----------
    length: int
        Number of elements in state vector.

    fitness_fn: fitness function object
        Object to implement fitness function for optimization.

    maximize: bool, default: True
        Whether to maximize the fitness function.
        Set :code:`False` for minimization problem.

    max_val: int, default: 2
        Number of unique values that each element in the state vector
        can take. Assumes values are integers in the range 0 to
        (max_val - 1), inclusive.
    """

    def __init__(self, length, fitness_fn, maximize=True, max_val=2, crossover=None, mutator=None):
        self._get_mutual_info_impl = self._get_mutual_info_slow

        OptProb.__init__(self, length, fitness_fn, maximize)

        if self.fitness_fn.get_problem_type() == "continuous":
            raise ValueError(
                """fitness_fn must have problem type 'discrete',"""
                + """ 'either' or 'tsp'. Define problem as"""
                + """ ContinuousOpt problem or use alternative"""
                + """ fitness function."""
            )

        if max_val < 0:
            raise ValueError(f"max_val must be a positive integer. Got {max_val}")
        elif not isinstance(max_val, int):
            if max_val.is_integer():
                self.max_val = int(max_val)
            else:
                raise ValueError(f"max_val must be a positive integer. Got {max_val}")
        else:
            self.max_val = max_val

        self.keep_sample = np.array([])
        self.node_probs = np.zeros([self.length, self.max_val, self.max_val])
        self.parent_nodes = np.array([])
        self.sample_order = []
        self.prob_type = "discrete"
        self.noise = 0

        self._crossover = UniformCrossover(self) if crossover is None else crossover
        self._mutator = GeneSwapMutator(self) if mutator is None else mutator

        self._mut_mask = None
        self._mut_inf = None

    def eval_node_probs(self):
        """Update probability density estimates."""
        # Create mutual info matrix
        mutual_info = self._get_mutual_info_impl()

        # Find minimum spanning tree of mutual info matrix
        csr_mx = csr_matrix(mutual_info)

        # noinspection PyTypeChecker
        mst = minimum_spanning_tree(csr_mx)

        # Convert minimum spanning tree to depth first tree with node 0 as root
        dft = depth_first_tree(csr_matrix(mst.toarray()), 0, directed=False)
        dft = np.round(dft.toarray(), 10)

        # Determine parent of each node
        parent = np.argmin(dft[:, 1:], axis=0)

        # Get probs
        probs = np.zeros([self.length, self.max_val, self.max_val])

        probs[0, :] = np.histogram(self.keep_sample[:, 0], np.arange(self.max_val + 1), density=True)[0]

        for i in range(1, self.length):
            for j in range(self.max_val):
                subset = self.keep_sample[np.where(self.keep_sample[:, parent[i - 1]] == j)[0]]

                if not len(subset):
                    probs[i, j] = 1 / self.max_val
                else:
                    temp_probs = np.histogram(subset[:, i], np.arange(self.max_val + 1), density=True)[0]

                    # Check if noise argument is not default (in epsilon)
                    if self.noise > 0:
                        # Add noise, from the mimic argument "noise"
                        temp_probs += self.noise
                        # All probability adds up to one
                        temp_probs = np.divide(temp_probs, np.sum(temp_probs))
                        # Handle floating point error to ensure probability adds up to 1
                        if sum(temp_probs) != 1.0:
                            temp_probs = np.divide(temp_probs, np.sum(temp_probs))
                    # Set probability
                    probs[i, j] = temp_probs

        # Update probs and parent
        self.node_probs = probs
        self.parent_nodes = parent

    def set_mimic_fast_mode(self, fast_mode):
        if fast_mode:
            mut_mask = np.zeros([self.length, self.length], dtype=bool)
            for i in range(0, self.length):
                for j in range(i, self.length):
                    mut_mask[i, j] = True
            mut_mask = mut_mask.reshape((self.length * self.length))
            self._mut_mask = mut_mask
            # Set ignore error to ignore dividing by zero
            np.seterr(divide="ignore", invalid="ignore")
            self._get_mutual_info_impl = self._get_mutual_info_fast
            self._mut_inf = np.zeros([self.length * self.length])
        else:
            self._mut_mask = None
            self._get_mutual_info_impl = self._get_mutual_info_slow
            self._mut_inf = None

    def _get_mutual_info_slow(self):
        mutual_info = np.zeros([self.length, self.length])
        for i in range(self.length - 1):
            for j in range(i + 1, self.length):
                mutual_info[i, j] = -1 * mutual_info_score(self.keep_sample[:, i], self.keep_sample[:, j])
        return mutual_info

    # adapted from https://github.com/parkds/mlrose/blob/f7154a1d3e3fdcd934bb3c683b943264d2870fd1/mlrose/algorithms.py
    # (thanks to David Sejin Park)
    def _get_mutual_info_fast(self):
        if self._mut_inf is None:
            # restore sanity
            self._get_mutual_info_impl = self._get_mutual_info_slow
            return self._get_mutual_info_impl()

        # get length of the sample which survived from mimic iteration
        len_sample_kept = self.keep_sample.shape[0]
        # get the length of the bit sequence / problem size
        len_prob = self.keep_sample.shape[1]

        # Expand the matrices to so each row corresponds to a row by row combination of the list of samples
        b = np.repeat(self.keep_sample, self.length).reshape(len_sample_kept, len_prob * len_prob)
        d = np.hstack(([self.keep_sample] * len_prob))

        # Compute the mutual information matrix in bulk, by iterating through the list of
        # possible feature values ((max_val-1)^2).
        # For example, a binary string would go through 00 01 10 11, for a total of 4 iterations.

        # First initialize the mutual info matrix.
        self._mut_inf.fill(0)
        # Pre-compute the U and V which gets computed multiple times in the inner loop.
        U = {}
        V = {}
        U_sum = {}
        V_sum = {}
        for i in range(0, self.max_val):
            U[i] = d == i
            V[i] = b == i
            U_sum[i] = np.sum(d == i, axis=0)
            V_sum[i] = np.sum(b == i, axis=0)

        # Compute the mutual information for all sample to sample combination for each feature combination
        # ((max_val-1)^2)
        for i in range(0, self.max_val):
            for j in range(0, self.max_val):
                # This corresponds to U and V of mutual info matrix, for this feature pair
                coeff = np.sum(U[i] * V[j], axis=0)
                # Compute length N, for the particular feature pair
                UV_length = U_sum[i] * V_sum[j]

                # compute the second term of the MI matrix
                temp = np.log(coeff) - np.log(UV_length) + np.log(len_sample_kept)
                # remove the nans and negative infinity
                temp[np.isnan(temp)] = 0
                temp[np.isneginf(temp)] = 0

                # combine the first and the second term, divide by the length N.
                # Add the whole MI matrix for the feature to the previously computed values
                div = temp * np.divide(coeff, len_sample_kept)
                div[self._mut_mask] = 0
                self._mut_inf += div

        # Need to multiply by negative to get the mutual information
        self._mut_inf = -self._mut_inf.reshape(self.length, self.length)
        # Only get the upper triangle matrix above the identity row.
        # Possible enhancements, currently we are doing double the computation required.
        # Pre set the matrix so the computation is only done for rows that are needed. To do for the future.

        mutual_info = self._mut_inf.T
        self._mut_inf = self._mut_inf.reshape(self.length * self.length)
        return mutual_info

    def find_neighbors(self):
        """Find all neighbors of the current state."""
        self.neighbors = []

        if self.max_val == 2:
            for i in range(self.length):
                neighbor = np.copy(self.state)
                neighbor[i] = np.abs(neighbor[i] - 1)
                self.neighbors.append(neighbor)

        else:
            for i in range(self.length):
                vals = list(np.arange(self.max_val))
                vals.remove(self.state[i])

                for j in vals:
                    neighbor = np.copy(self.state)
                    neighbor[i] = j
                    self.neighbors.append(neighbor)

    def find_sample_order(self):
        """Determine order in which to generate sample vector elements."""
        sample_order = []
        last = [0]
        parent = self.parent_nodes

        while len(sample_order) < self.length:
            inds = []

            # If last nodes list is empty, select random node than has not
            # previously been selected
            if len(last) == 0:
                inds = [np.random.choice(list(set(np.arange(self.length)) - set(sample_order)))]
            else:
                for i in last:
                    inds += list(np.where(parent == i)[0] + 1)

            sample_order += last
            last = inds

        self.sample_order = sample_order

    def find_top_pct(self, keep_pct):
        """Select samples with fitness in the top keep_pct percentile.

        Parameters
        ----------
        keep_pct: float
            Proportion of samples to keep.
        """
        if (keep_pct < 0) or (keep_pct > 1):
            raise Exception("""keep_pct must be between 0 and 1.""")

        # Determine threshold
        theta = np.percentile(self.pop_fitness, 100 * (1 - keep_pct))

        # Determine samples for keeping
        keep_inds = np.where(self.pop_fitness >= theta)[0]

        # Determine sample for keeping
        self.keep_sample = self.population[keep_inds]

    def get_keep_sample(self):
        """Return the keep sample.

        Returns
        -------
        self.keep_sample: np.ndarray
            Numpy array containing samples with fitness in the top keep_pct
            percentile.
        """
        return self.keep_sample

    def get_problem_type(self):
        """Return the problem type.

        Returns
        -------
        self.prob_type: str
            Returns problem type.
        """
        return self.prob_type

    def random(self):
        """Return a random state vector.

        Returns
        -------
        state: np.ndarray
            Randomly generated state vector.
        """
        state = np.random.randint(0, self.max_val, self.length)

        return state

    def random_neighbor(self):
        """Return random neighbor of current state vector.

        Returns
        -------
        neighbor: np.ndarray
            State vector of random neighbor.
        """
        neighbor = np.copy(self.state)
        i = np.random.randint(0, self.length)

        if self.max_val == 2:
            neighbor[i] = np.abs(neighbor[i] - 1)

        else:
            vals = list(np.arange(self.max_val))
            vals.remove(neighbor[i])
            neighbor[i] = vals[np.random.randint(0, self.max_val - 1)]

        return neighbor

    def random_pop(self, pop_size):
        """Create a population of random state vectors.

        Parameters
        ----------
        pop_size: int
            Size of population to be created.
        """
        if pop_size <= 0:
            raise Exception("""pop_size must be a positive integer.""")
        elif not isinstance(pop_size, int):
            if pop_size.is_integer():
                pop_size = int(pop_size)
            else:
                raise Exception("""pop_size must be a positive integer.""")

        population = []
        pop_fitness = []

        for _ in range(pop_size):
            state = self.random()
            population.append(state)
            fitness = self.eval_fitness(state)
            pop_fitness.append(fitness)

        self.population = np.array(population)
        self.pop_fitness = np.array(pop_fitness)

    def reproduce(self, parent_1, parent_2, mutation_prob=0.1):
        """Create child state vector from two parent state vectors.

        Parameters
        ----------
        parent_1: np.ndarray
            State vector for parent 1.
        parent_2: np.ndarray
            State vector for parent 2.
        mutation_prob: float
            Probability of a mutation at each state element during
            reproduction.

        Returns
        -------
        child: np.ndarray
            Child state vector produced from parents 1 and 2.
        """
        if len(parent_1) != self.length or len(parent_2) != self.length:
            raise Exception("""Lengths of parents must match problem length""")

        if (mutation_prob < 0) or (mutation_prob > 1):
            raise Exception("""mutation_prob must be between 0 and 1.""")

        # Reproduce parents
        child = self._crossover.mate(parent_1, parent_2)

        # Mutate child
        child = self._mutator.mutate(child, mutation_prob)

        return child

    def reset(self):
        """Set the current state vector to a random value and get its fitness."""
        self.state = self.random()
        self.fitness = self.eval_fitness(self.state)
        self.fevals = {}
        self.fitness_evaluations = 0
        self.current_iteration = 0

    def sample_pop(self, sample_size):
        """Generate new sample from probability density.

        Parameters
        ----------
        sample_size: int
            Size of sample to be generated.

        Returns
        -------
        new_sample: np.ndarray
            Numpy array containing new sample.
        """
        if sample_size <= 0:
            raise Exception("""sample_size must be a positive integer.""")
        elif not isinstance(sample_size, int):
            if sample_size.is_integer():
                sample_size = int(sample_size)
            else:
                raise Exception("""sample_size must be a positive integer.""")

        # Initialize new sample matrix
        new_sample = np.zeros([sample_size, self.length])

        # Get value of first element in new samples
        new_sample[:, 0] = np.random.choice(self.max_val, sample_size, p=self.node_probs[0, 0])

        # Get sample order
        self.find_sample_order()
        sample_order = self.sample_order[1:]

        # Get values for remaining elements in new samples
        for i in sample_order:
            par_ind = self.parent_nodes[i - 1]

            for j in range(self.max_val):
                inds = np.where(new_sample[:, par_ind] == j)[0]
                new_sample[inds, i] = np.random.choice(self.max_val, len(inds), p=self.node_probs[i, j])

        return new_sample



import os
import time
import hashlib
import logging
from abc import ABC
import numpy as np
import pandas as pd
import pickle as pk

from mlrose_ky import GridSearchMixin
from mlrose_ky.runners._runner_base import _RunnerBase


class _NNRunnerBase(_RunnerBase, GridSearchMixin, ABC):
    _interrupted_result_list = []

    def __init__(
        self,
        x_train,
        y_train,
        x_test,
        y_test,
        experiment_name,
        seed,
        iteration_list,
        grid_search_parameters,
        grid_search_scorer_method,
        cv=5,
        generate_curves=True,
        output_directory=None,
        verbose_grid_search=True,
        override_ctrl_c_handler=True,
        n_jobs=1,
        replay=False,
        **kwargs,
    ):
        # call super on _RunnerBase
        _RunnerBase.__init__(
            self,
            problem=None,
            experiment_name=experiment_name,
            seed=seed,
            iteration_list=iteration_list,
            generate_curves=generate_curves,
            output_directory=output_directory,
            replay=replay,
            override_ctrl_c_handler=override_ctrl_c_handler,
            copy_zero_curve_fitness_from_first=True,
        )

        # call super on GridSearchMixin
        GridSearchMixin.__init__(self, scorer_method=grid_search_scorer_method)

        self.classifier = None

        # add algorithm grid-search params
        self.grid_search_parameters = self.build_grid_search_parameters(grid_search_parameters=grid_search_parameters, **kwargs)
        self.x_train = x_train
        self.y_train = y_train
        self.x_test = x_test
        self.y_test = y_test
        self.cv = cv
        self.n_jobs = n_jobs
        self.verbose_grid_search = verbose_grid_search
        self.cv_results_df = None
        self.best_params = None

    def run(self):
        try:
            self._setup()
            logging.info(f"Running {self.dynamic_runner_name()}")
            if self.replay_mode():
                gsr_name = f"{super()._get_pickle_filename_root('grid_search_results')}.p"
                with open(gsr_name, "rb") as pickle_file:
                    sr = pk.load(pickle_file)
            else:
                run_start = time.perf_counter()
                sr = self._perform_grid_search(
                    classifier=self.classifier,
                    parameters=self.grid_search_parameters,
                    x_train=self.x_train,
                    y_train=self.y_train,
                    cv=self.cv,
                    n_jobs=self.n_jobs,
                    verbose=self.verbose_grid_search,
                )
                run_end = time.perf_counter()
                logging.info(f"Run time: {run_end - run_start}")

            # pull the stats from the best estimator to here.
            # (as grid search will have cloned this object).
            self.__dict__.update(sr.best_estimator_.runner.__dict__)

            self.best_params = sr.best_params_
            # dump the results to disk
            self.cv_results_df = self._make_cv_results_data_frame(sr.cv_results_)
            edf = {"cv_results_df": self.cv_results_df}
            self._create_and_save_run_data_frames(extra_data_frames=edf, final_save=True)

            try:
                self._dump_pickle_to_disk(sr, "grid_search_results", final_save=True)
            except (OSError, IOError, pk.PickleError):
                pass

            # noinspection PyBroadException
            try:
                y_pred = sr.best_estimator_.predict(self.x_test)
                score = self.score(y_pred=y_pred, y_true=self.y_train)
                self._print_banner(f"Score: {score}")
            except Exception:
                pass

            return self.run_stats_df, self.curves_df, self.cv_results_df, sr
        except KeyboardInterrupt:
            return None, None, None, None
        finally:
            self._tear_down()

    def _get_pickle_filename_root(self, name):
        filename_root = super()._get_pickle_filename_root(name)
        arg_text = "".join([f"{k}_{self._sanitize_value(v)}_" for k, v in self._current_logged_algorithm_args.items() if "state" not in k])
        arg_hash = f"__{hashlib.md5(arg_text.encode()).hexdigest()}".upper() if len(arg_text) > 0 else ""
        filename_root += arg_hash
        return filename_root

    @staticmethod
    def _check_match(df_ref, df_to_check):
        cols = [i for i in df_ref.columns]
        found = False
        for _, row in df_to_check.iterrows():
            found = True
            for col in cols:
                if df_ref[col][0] != row[col]:
                    found = False
                    break
            if found:
                break
        return found

    def _tear_down(self, filename=None):
        if self.best_params is None or self.replay_mode() is None or self._output_directory is None:
            super()._tear_down()
            return

        filename_root = super()._get_pickle_filename_root("")

        path = os.path.join(*filename_root.split(os.sep)[:-1])
        filename_part = filename_root.split(os.sep)[-1]
        if not os.path.isdir(path) and path[0] != os.sep:
            path = f"{os.sep}{path}"

        # find all data frames output by this runner
        filenames = [fn for fn in os.listdir(str(path)) if (filename_part in fn and fn.endswith(".p") and "_df_" in fn)]

        # get the best parameters
        df_best_params = pd.DataFrame([{k: self._sanitize_value(v) for k, v in self.best_params.items()}])

        # file the files that match the best parameters (and don't)
        correct_files = []
        incorrect_files = []
        for fn in filenames:
            filename = os.path.join(str(path), fn)
            with open(filename, "rb") as pickle_file:
                try:
                    df = pk.load(pickle_file)
                    found = self._check_match(df_best_params, df)
                    if not found:
                        incorrect_files.append(filename)
                    else:
                        correct_files.append(filename)
                except (EOFError, pk.UnpicklingError):
                    pass

        # extract the md5s from the names for the best and non-best parameter files
        correct_md5s = list(set([p.split("_")[-1][:-2] for p in correct_files]))
        incorrect_md5s = list(set([p.split("_")[-1][:-2] for p in incorrect_files]))

        # remove the suboptimal files
        all_incorrect_files = []
        for incorrect_md5 in incorrect_md5s:
            all_incorrect_files.extend([os.path.join(str(path), fn) for fn in os.listdir(str(path)) if incorrect_md5 in fn])

        for _filename in all_incorrect_files:
            # os.rename(filename, f'{filename}.del')
            os.remove(_filename)

        # rename the best files by removing the md5 from the end
        all_correct_files = []
        for _correct_md5 in correct_md5s:
            all_correct_files.extend(
                [(os.path.join(str(path), fn), f"__{_correct_md5}") for fn in os.listdir(str(path)) if _correct_md5 in fn]
            )

        for _filename, _correct_md5 in all_correct_files:
            correct_filename = _filename.replace(_correct_md5, "")
            if os.path.exists(correct_filename):
                os.rename(correct_filename, f"{correct_filename}.bak")
            os.rename(_filename, correct_filename)

        super()._tear_down()

    @staticmethod
    def _make_cv_results_data_frame(cv_results):
        cv_results = cv_results.copy()
        param_prefix = "param_"
        # drop params
        param_labels = [k for k in cv_results if param_prefix in k]
        # clean_results = {k: v for k, v in cv_results.items() if 'param_' not in k}

        new_param_values = {p: [] for p in param_labels}
        for v in cv_results["params"]:
            for p in param_labels:
                pl = p.replace(param_prefix, "")
                new_param_values[p].append(_NNRunnerBase._sanitize_value(v[pl]))

        # replace values with sanitized values
        cv_results.update(new_param_values)
        df = pd.DataFrame(cv_results)
        df.dropna(inplace=True)
        return df

    @staticmethod
    def build_grid_search_parameters(grid_search_parameters, **kwargs):
        # extract nn parameters
        all_grid_search_parameters = {
            # 'hidden_nodes': hidden_nodes_set,
            # 'activation': activation_set,
            # 'learning_rate': learning_rates
        }
        all_grid_search_parameters.update(grid_search_parameters)
        all_grid_search_parameters.update(**kwargs)
        return all_grid_search_parameters

    def _grid_search_score_intercept(self, y_true, y_pred, sample_weight=None, adjusted=False):
        if not self.classifier.fit_started_ and self.has_aborted():
            return np.NaN
        return super()._grid_search_score_intercept(y_true=y_true, y_pred=y_pred, sample_weight=sample_weight, adjusted=adjusted)



"""Class defining the Knapsack fitness function for use with optimization algorithms."""

# Authors: Genevieve Hayes (modified by Andrew Rollings, Kyle Nakamura)
# License: BSD 3 clause

import numpy as np


class Knapsack:
    """Fitness function for Knapsack optimization problem.

    Given a set of n items, where item i has known weight :math:`w_{i}` and known value
    :math:`v_{i}`; and maximum knapsack capacity, :math:`W`, the Knapsack fitness function
    evaluates the fitness of a state vector :math:`x = [x_{0}, x_{1}, \\ldots, x_{n-1}]` as:

    .. math::

        Fitness(x) = \\sum_{i = 0}^{n-1}v_{i}x_{i}, \\text{ if}
        \\sum_{i = 0}^{n-1}w_{i}x_{i} \\leq W, \\text{ and 0, otherwise,}

    where :math:`x_{i}` denotes the number of copies of item i included in the knapsack.

    Parameters
    ----------
    weights : list[float]
        List of weights for each of the n items.

    values : list[float]
        List of values for each of the n items.

    max_weight_pct : float, default=0.35
        Parameter used to set maximum capacity of knapsack (W) as a percentage of
        the total of the weights list (:math:`W =` max_weight_pct :math:`\times` total_weight).

    max_item_count : int, default=1
        Maximum number of copies of each item that can be included in the knapsack.

    multiply_by_max_item_count : bool, default=False
        Whether to multiply the maximum weight by the maximum item count.

    Examples
    --------
    >>> weights = [10, 5, 2, 8, 15]
    >>> values = [1, 2, 3, 4, 5]
    >>> max_weight_pct = 0.6
    >>> fitness = Knapsack(weights, values, max_weight_pct)
    >>> state_vector = np.array([1, 0, 2, 1, 0])
    >>> fitness.evaluate(state_vector)
    11.0

    Note
    ----
    The Knapsack fitness function is suitable for use in discrete-state optimization problems *only*.
    """

    def __init__(
        self,
        weights: list[float],
        values: list[float],
        max_weight_pct: float = 0.35,
        max_item_count: int = 1,
        multiply_by_max_item_count: bool = False,
    ):
        self.problem_type: str = "discrete"
        self.weights: list[float] = weights
        self.values: list[float] = values

        count_multiplier = max_item_count if multiply_by_max_item_count else 1.0
        self.max_weight = np.ceil(np.sum(self.weights) * max_weight_pct * count_multiplier)

        if len(self.weights) != len(self.values):
            raise ValueError("The weights and values lists must be the same size.")
        if min(self.weights) <= 0:
            raise ValueError("All weights must be greater than 0.")
        if min(self.values) <= 0:
            raise ValueError("All values must be greater than 0.")
        if max_item_count <= 0:
            raise ValueError("max_item_count must be greater than 0.")
        if max_weight_pct <= 0:
            raise ValueError("max_weight_pct must be greater than 0.")

    def evaluate(self, state_vector: np.ndarray) -> float:
        """Evaluate the fitness of a state vector.

        Parameters
        ----------
        state_vector : np.ndarray
            State array for evaluation. Must be the same length as the weights
            and values arrays.

        Returns
        -------
        float
            Value of fitness function.

        Raises
        ------
        ValueError
            If `state_vector` is not the same size as the weights and values arrays.
        TypeError
            If `state_vector` is not an instance of `np.ndarray`.
        """
        if not isinstance(state_vector, np.ndarray):
            raise TypeError(f"Expected state_vector to be np.ndarray, got {type(state_vector).__name__} instead.")
        if len(state_vector) != len(self.weights):
            raise ValueError("The state_vector must be the same size as the weights and values arrays.")

        total_weight = np.sum(state_vector * self.weights)
        total_value = np.sum(state_vector * self.values)

        if total_weight <= self.max_weight:
            return float(total_value)

        return 0.0

    def get_problem_type(self) -> str:
        """Return the problem type.

        Returns
        -------
        str
            Specifies problem type as 'discrete'.
        """
        return self.problem_type



"""Discrete Gene Mutator for Genetic Algorithms (GA)."""

# Authors: Genevieve Hayes (modified by Andrew Rollings, Kyle Nakamura)
# License: BSD 3 clause

import numpy as np
from typing import Any

from mlrose_ky.algorithms.mutators._mutator_base import _MutatorBase


class DiscreteGeneMutator(_MutatorBase):
    """
    A mutator class that performs discrete mutation on individual genes in a genetic algorithm.

    This class supports mutation where each gene in the chromosome can be discretely modified
    based on the mutation probability and the nature of the gene values.

    Attributes
    ----------
    optimization_problem : Any
        The optimization problem instance associated with the mutation operations.
    max_gene_value : int
        The maximum allowable value for any gene in the chromosome.

    Parameters
    ----------
    optimization_problem : Any
        An instance of an optimization problem that the mutator will operate on.
    """

    def __init__(self, optimization_problem: Any) -> None:
        super().__init__(optimization_problem)
        self.max_gene_value = optimization_problem.max_val

    def mutate(self, child: np.ndarray, mutation_probability: float) -> np.ndarray:
        """
        Apply discrete mutation to a chromosome based on a given mutation probability.

        The method supports two mutation modes: binary mutation (for binary genes) and discrete
        value mutation for genes that can take on multiple discrete values.

        Parameters
        ----------
        child : np.ndarray
            The chromosome of a child individual to be mutated.
        mutation_probability : float
            The probability of each gene being mutated.

        Returns
        -------
        np.ndarray
            The mutated chromosome.
        """
        random_thresholds = np.random.uniform(size=self.chromosome_length)
        mutation_indices = np.where(random_thresholds < mutation_probability)[0]

        if self.max_gene_value == 2:
            child[mutation_indices] = 1 - child[mutation_indices]
        else:
            for index in mutation_indices:
                possible_values = list(range(self.max_gene_value))
                possible_values.remove(child[index])
                child[index] = np.random.choice(possible_values)

        return child



import mlrose_ky
from mlrose_ky.decorators import short_name
from mlrose_ky.runners._runner_base import _RunnerBase

"""
Example usage:

    experiment_name = 'example_experiment'
    problem = TSPGenerator.generate(seed=SEED, number_of_cities=22)

    mmc = MIMICRunner(problem=problem,
                      experiment_name=experiment_name,
                      output_directory=OUTPUT_DIRECTORY,
                      seed=SEED,
                      iteration_list=2 ** np.arange(10),
                      max_attempts=500,
                      keep_percent_list=[0.25, 0.5, 0.75])
                      
    # the two data frames will contain the results
    df_run_stats, df_run_curves = mmc.run()
"""


@short_name("mimic")
class MIMICRunner(_RunnerBase):

    def __init__(
        self,
        problem,
        experiment_name,
        seed,
        iteration_list,
        population_sizes,
        keep_percent_list,
        max_attempts=5,
        generate_curves=True,
        use_fast_mimic=True,
        **kwargs,
    ):
        super().__init__(
            problem=problem,
            experiment_name=experiment_name,
            seed=seed,
            iteration_list=iteration_list,
            max_attempts=max_attempts,
            generate_curves=generate_curves,
            **kwargs,
        )
        self.keep_percent_list = keep_percent_list
        self.population_sizes = population_sizes
        self._use_fast_mimic = None
        if hasattr(problem, "set_mimic_fast_mode") and callable(getattr(problem, "set_mimic_fast_mode")):
            self._use_fast_mimic = use_fast_mimic
            problem.set_mimic_fast_mode(use_fast_mimic)

    def _setup(self):
        super()._setup()
        if self._use_fast_mimic is not None:
            self.log_current_argument("use_fast_mimic", self._use_fast_mimic)

    def run(self):
        return super().run_experiment_(
            algorithm=mlrose_ky.mimic,
            pop_size=("Population Size", self.population_sizes),
            keep_pct=("Keep Percent", self.keep_percent_list),
        )
